@@ -46,18 +46,17 @@ func (p *UcpRequestParams) SetCallback(cb UcpCallback) *UcpRequestParams {
 	return p
 }
 
-func packParams(params *UcpRequestParams, p *C.ucp_request_param_t, cb unsafe.Pointer) uint64 {
+func packParams(params *UcpRequestParams, p *C.ucp_request_param_t, cb unsafe.Pointer) unsafe.Pointer {
+
 	if params == nil {
-		return 0
+		return nil
 	}
 
-	var cbId uint64
 	if params.Cb != nil {
-		cbId = register(params.Cb)
 		p.op_attr_mask |= C.UCP_OP_ATTR_FIELD_CALLBACK | C.UCP_OP_ATTR_FIELD_USER_DATA
 		cbAddr := (*unsafe.Pointer)(unsafe.Pointer(&p.cb[0]))
 		*cbAddr = cb
-		p.user_data = unsafe.Pointer(uintptr(cbId))
+		p.user_data = unsafe.Pointer(packCallback(params.Cb))
 	}
 
 	if params.memTypeSet {
@@ -74,7 +73,7 @@ func packParams(params *UcpRequestParams, p *C.ucp_request_param_t, cb unsafe.Po
 		p.memh = params.Memory.memHandle
 	}
 
-	return cbId
+	return p.user_data
 }
 
 // Checks whether request is a pointer
@@ -83,7 +82,7 @@ func isRequestPtr(request C.ucs_status_ptr_t) bool {
 	return (uint64(uintptr(request)) - 1) < (uint64(errLast) - 1)
 }
 
-func NewRequest(request C.ucs_status_ptr_t, callbackId uint64, immidiateInfo interface{}) (*UcpRequest, error) {
+func newRequest(request C.ucs_status_ptr_t, packedCb unsafe.Pointer, immediateInfo interface{}) (*UcpRequest, error) {
 	ucpRequest := &UcpRequest{}
 
 	if isRequestPtr(request) {
@@ -91,14 +90,14 @@ func NewRequest(request C.ucs_status_ptr_t, callbackId uint64, immidiateInfo int
 		ucpRequest.Status = UCS_INPROGRESS
 	} else {
 		ucpRequest.Status = UcsStatus(int64(uintptr(request)))
-		if callback, found := deregister(callbackId); found {
+		if callback := PackedCallback(packedCb).unpackAndFree(); callback != nil {
 			switch callback := callback.(type) {
 			case UcpSendCallback:
 				callback(ucpRequest, ucpRequest.Status)
 			case UcpTagRecvCallback:
-				callback(ucpRequest, ucpRequest.Status, immidiateInfo.(*UcpTagRecvInfo))
+				callback(ucpRequest, ucpRequest.Status, immediateInfo.(*UcpTagRecvInfo))
 			case UcpAmDataRecvCallback:
-				callback(ucpRequest, ucpRequest.Status, uint64(immidiateInfo.(C.size_t)))
+				callback(ucpRequest, ucpRequest.Status, uint64(immediateInfo.(C.size_t)))
 			}
 		}
 		if ucpRequest.Status != UCS_OK {
