@@ -28,19 +28,19 @@ public class UcpEndpointTest extends UcxTest {
 
     @DataPoints
     public static ArrayList<Integer> memTypes() {
-        ArrayList<Integer> resut = new ArrayList<>();
-        resut.add(UcsConstants.MEMORY_TYPE.UCS_MEMORY_TYPE_HOST);
+        ArrayList<Integer> result = new ArrayList<>();
+        result.add(UcsConstants.MEMORY_TYPE.UCS_MEMORY_TYPE_HOST);
         UcpContext testContext = new UcpContext(new UcpParams().requestTagFeature());
         long memTypeMask = testContext.getMemoryTypesMask();
         if (UcsConstants.MEMORY_TYPE.isMemTypeSupported(memTypeMask,
             UcsConstants.MEMORY_TYPE.UCS_MEMORY_TYPE_CUDA)) {
-            resut.add(UcsConstants.MEMORY_TYPE.UCS_MEMORY_TYPE_CUDA);
+            result.add(UcsConstants.MEMORY_TYPE.UCS_MEMORY_TYPE_CUDA);
         }
         if (UcsConstants.MEMORY_TYPE.isMemTypeSupported(memTypeMask,
             UcsConstants.MEMORY_TYPE.UCS_MEMORY_TYPE_CUDA_MANAGED)) {
-            resut.add(UcsConstants.MEMORY_TYPE.UCS_MEMORY_TYPE_CUDA_MANAGED);
+            result.add(UcsConstants.MEMORY_TYPE.UCS_MEMORY_TYPE_CUDA_MANAGED);
         }
-        return resut;
+        return result;
     }
 
     @Test
@@ -246,10 +246,12 @@ public class UcpEndpointTest extends UcxTest {
         UcpEndpoint ep = worker1.newEndpoint(new UcpEndpointParams().setName("testSendRecv")
             .setUcpAddress(worker2.getAddress()));
 
-        ep.sendTaggedNonBlocking(src1.getMemory().getAddress(), UcpMemoryTest.MEM_SIZE, 0, null,
-            new UcpRequestParams().setMemoryType(memType).setMemoryHandle(src1.getMemory()));
-        ep.sendTaggedNonBlocking(src2.getMemory().getAddress(), UcpMemoryTest.MEM_SIZE, 1, null,
-            new UcpRequestParams().setMemoryType(memType).setMemoryHandle(src2.getMemory()));
+        UcpRequest send1 = ep.sendTaggedNonBlocking(src1.getMemory().getAddress(),
+                UcpMemoryTest.MEM_SIZE, 0, null,
+                new UcpRequestParams().setMemoryType(memType).setMemoryHandle(src1.getMemory()));
+        UcpRequest send2 = ep.sendTaggedNonBlocking(src2.getMemory().getAddress(),
+                UcpMemoryTest.MEM_SIZE, 1, null,
+                new UcpRequestParams().setMemoryType(memType).setMemoryHandle(src2.getMemory()));
 
         while (receivedMessages.get() != 2) {
             worker1.progress();
@@ -259,13 +261,16 @@ public class UcpEndpointTest extends UcxTest {
         assertEquals(src1.getData().asCharBuffer(), dst1.getData().asCharBuffer());
         assertEquals(src2.getData().asCharBuffer(), dst2.getData().asCharBuffer());
 
+        worker1.progressRequest(send1);
+        worker1.progressRequest(send2);
+
         Collections.addAll(resources, context2, context1, worker2, worker1, ep,
             src1, src2, dst1, dst2);
         closeResources();
     }
 
     @Test
-    public void testRecvAfterSend() {
+    public void testRecvAfterSend() throws Exception {
         long sendTag = 4L;
         // Create 2 contexts + 2 workers
         UcpParams params = new UcpParams().requestRmaFeature().requestTagFeature()
@@ -285,7 +290,7 @@ public class UcpEndpointTest extends UcxTest {
         ByteBuffer src1 = ByteBuffer.allocateDirect(UcpMemoryTest.MEM_SIZE);
         ByteBuffer dst1 = ByteBuffer.allocateDirect(UcpMemoryTest.MEM_SIZE);
 
-        ep.sendTaggedNonBlocking(src1, sendTag, null);
+        UcpRequest send = ep.sendTaggedNonBlocking(src1, sendTag, null);
 
         Thread progressThread = new Thread() {
             @Override
@@ -325,6 +330,8 @@ public class UcpEndpointTest extends UcxTest {
 
         assertTrue(recv.isCompleted());
         assertEquals(sendTag, recv.getSenderTag());
+        worker1.progressRequest(send);
+
         UcpRequest closeRequest = ep.closeNonBlockingForce();
 
         while (!closeRequest.isCompleted()) {
@@ -571,10 +578,11 @@ public class UcpEndpointTest extends UcxTest {
             recvAddresses[i] = recvBuffers[i].getAddress();
         }
 
-        ep.sendTaggedNonBlocking(sendAddresses, sizes, 0L, null);
-        UcpRequest recv = worker2.recvTaggedNonBlocking(recvAddresses, sizes, 0L, 0L, null);
+        UcpRequest send = ep.sendTaggedNonBlocking(sendAddresses, sizes, 0L, null);
+        UcpRequest recv = worker2.recvTaggedNonBlocking(recvAddresses, sizes, 0L, 0L,
+                null);
 
-        while (!recv.isCompleted()) {
+        while (!recv.isCompleted() || !send.isCompleted()) {
             worker1.progress();
             worker2.progress();
         }
@@ -659,7 +667,7 @@ public class UcpEndpointTest extends UcxTest {
 
         AtomicBoolean errorCallabackCalled = new AtomicBoolean(false);
 
-        ep.sendTaggedNonBlocking(src, null);
+        send = ep.sendTaggedNonBlocking(src, null);
         worker1.progressRequest(ep.flushNonBlocking(new UcxCallback() {
             @Override
             public void onError(int ucsStatus, String errorMsg) {
@@ -667,6 +675,7 @@ public class UcpEndpointTest extends UcxTest {
             }
         }));
 
+        assertTrue(send.isCompleted());
         assertTrue(errorHandlerCalled.get());
         assertTrue(errorCallabackCalled.get());
 
