@@ -11,6 +11,7 @@
 #define UCX_LIBPERF_H
 
 #include <ucs/sys/compiler.h>
+#include <ucs/sys/device_code.h>
 
 BEGIN_C_DECLS
 
@@ -30,6 +31,9 @@ typedef enum {
 typedef enum {
     UCX_PERF_CMD_AM,
     UCX_PERF_CMD_PUT,
+    UCX_PERF_CMD_PUT_SINGLE,
+    UCX_PERF_CMD_PUT_MULTI,
+    UCX_PERF_CMD_PUT_PARTIAL,
     UCX_PERF_CMD_GET,
     UCX_PERF_CMD_ADD,
     UCX_PERF_CMD_FADD,
@@ -118,8 +122,6 @@ enum {
  *      |                 |---------PEER_TX(AM)------->|                 |
  *      |                 |----------(TM msg)--------->|                 |
  *      |<---SEND_CMPL----|                            |---RECV_CMPL---->|
- *      |                 |                            |                 |
- *      |-------FIN------>|                            |<------FIN-------|
  *
  * (AM) used only when AM communication is configured
  * (TM) used only when tag matching communication is configured
@@ -144,11 +146,9 @@ enum {
      * unsolicited in case of AM communication. For tag matching transport it \
      * confirms the completion of RECV_REQ request */ \
     _macro(UCP_PERF_DAEMON_AM_ID_RECV_CMPL, UCP_PERF_AM_ID + 5) \
-    /* Host->DPU daemon finalize message */ \
-    _macro(UCP_PERF_DAEMON_AM_ID_FIN,       UCP_PERF_AM_ID + 6) \
     /* Sender DPU->Receiver DPU, message to establish communication between \
      * DPU peers */ \
-    _macro(UCP_PERF_DAEMON_AM_ID_PEER_INIT, UCP_PERF_AM_ID + 7)
+    _macro(UCP_PERF_DAEMON_AM_ID_PEER_INIT, UCP_PERF_AM_ID + 6)
 
 
 #define UCP_PERF_DAEMON_ENUMIFY(ID, VALUE) ID = VALUE,
@@ -164,6 +164,9 @@ typedef enum {
                                              (_params)->uct.dev_name
 
 
+#define UCX_PERF_MEM_DEV_DEFAULT -1
+
+
 /**
  * Performance counter type.
  */
@@ -189,6 +192,12 @@ typedef struct ucx_perf_result {
 } ucx_perf_result_t;
 
 
+typedef struct {
+    ucs_memory_type_t mem_type;
+    int               device_id;
+} ucx_perf_accel_dev_t;
+
+
 typedef void (*ucx_perf_rte_progress_cb_t)(void *arg);
 
 typedef ucs_status_t (*ucx_perf_rte_setup_func_t)(void *arg);
@@ -257,6 +266,9 @@ typedef struct ucx_perf_params {
     ucx_perf_wait_mode_t   wait_mode;       /* How to wait */
     ucs_memory_type_t      send_mem_type;   /* Send memory type */
     ucs_memory_type_t      recv_mem_type;   /* Recv memory type */
+    ucx_perf_accel_dev_t   send_device;     /* Send memory device for gdaki */
+    ucx_perf_accel_dev_t   recv_device;     /* Recv memory device for gdaki */
+    ucs_device_level_t     device_level;    /* Device level for gdaki */
     unsigned               flags;           /* See ucx_perf_test_flags. */
 
     size_t                 *msg_size_list;  /* Test message sizes list. The size
@@ -275,6 +287,8 @@ typedef struct ucx_perf_params {
     double                 report_interval; /* Interval at which to call the report callback */
     double                 percentile_rank; /* The percentile rank of the percentile reported
                                                in latency tests */
+    unsigned               device_thread_count; /* Number of device threads */
+    unsigned               device_block_count; /* Number of device blocks */
 
     void                   *rte_group;      /* Opaque RTE group handle */
     ucx_perf_rte_t         *rte;            /* RTE functions used to exchange data */
@@ -318,6 +332,10 @@ typedef struct {
 typedef struct ucx_perf_allocator ucx_perf_allocator_t;
 extern const ucx_perf_allocator_t* ucx_perf_mem_type_allocators[];
 
+/* Device test dispatchers for each memory type */
+typedef struct ucx_perf_device_dispatcher ucx_perf_device_dispatcher_t;
+extern const ucx_perf_device_dispatcher_t* ucx_perf_mem_type_device_dispatchers[];
+
 
 const char *ucp_perf_daemon_am_id_name(ucp_perf_daemon_am_id_t id);
 
