@@ -54,7 +54,10 @@ test_type_t tests[] = {
      "active message bandwidth / message rate", "overhead", 1},
 
     {"put_bw", UCX_PERF_API_UCT, UCX_PERF_CMD_PUT, UCX_PERF_TEST_TYPE_STREAM_UNI,
-     "put bandwidth / message rate", "overhead", 1},
+     "put bandwidth / message rate", "overhead", 32},
+
+    {"get_bw", UCX_PERF_API_UCT, UCX_PERF_CMD_GET, UCX_PERF_TEST_TYPE_STREAM_UNI,
+     "get bandwidth / message rate", "overhead", 32},
 
     {"add_mr", UCX_PERF_API_UCT, UCX_PERF_CMD_ADD, UCX_PERF_TEST_TYPE_STREAM_UNI,
      "atomic add message rate", "overhead", 1},
@@ -77,6 +80,24 @@ test_type_t tests[] = {
     {"ucp_put_bw", UCX_PERF_API_UCP, UCX_PERF_CMD_PUT, UCX_PERF_TEST_TYPE_STREAM_UNI,
      "put bandwidth", "overhead", 32},
 
+    {"ucp_put_single_bw", UCX_PERF_API_UCP, UCX_PERF_CMD_PUT_SINGLE, UCX_PERF_TEST_TYPE_STREAM_UNI,
+     "put single bandwidth", "overhead", 32},
+
+    {"ucp_put_single_lat", UCX_PERF_API_UCP, UCX_PERF_CMD_PUT_SINGLE, UCX_PERF_TEST_TYPE_PINGPONG,
+     "put single latency", "latency", 1},
+
+    {"ucp_put_multi_bw", UCX_PERF_API_UCP, UCX_PERF_CMD_PUT_MULTI, UCX_PERF_TEST_TYPE_STREAM_UNI,
+     "put multi bandwidth", "overhead", 32},
+
+    {"ucp_put_multi_lat", UCX_PERF_API_UCP, UCX_PERF_CMD_PUT_MULTI, UCX_PERF_TEST_TYPE_PINGPONG,
+     "put multi latency", "latency", 32},
+
+    {"ucp_put_partial_bw", UCX_PERF_API_UCP, UCX_PERF_CMD_PUT_PARTIAL, UCX_PERF_TEST_TYPE_STREAM_UNI,
+     "put partial bandwidth", "overhead", 32},
+
+    {"ucp_put_partial_lat", UCX_PERF_API_UCP, UCX_PERF_CMD_PUT_PARTIAL, UCX_PERF_TEST_TYPE_PINGPONG,
+     "put partial latency", "latency", 32},
+
     {"ucp_get", UCX_PERF_API_UCP, UCX_PERF_CMD_GET, UCX_PERF_TEST_TYPE_STREAM_UNI,
      "get latency / bandwidth / message rate", "latency", 1},
 
@@ -167,12 +188,14 @@ static int safe_recv(int sock, void *data, size_t size,
 ucs_status_t init_test_params(perftest_params_t *params)
 {
     static const struct sockaddr_storage empty_addr = {};
+    static const ucx_perf_accel_dev_t default_dev   =
+                            {UCS_MEMORY_TYPE_LAST, UCX_PERF_MEM_DEV_DEFAULT};
 
     memset(params, 0, sizeof(*params));
     params->super.api               = UCX_PERF_API_LAST;
     params->super.command           = UCX_PERF_CMD_LAST;
     params->super.test_type         = UCX_PERF_TEST_TYPE_LAST;
-    params->super.thread_mode       = UCS_THREAD_MODE_SINGLE;
+    params->super.thread_mode       = UCS_THREAD_MODE_SERIALIZED;
     params->super.thread_count      = 1;
     params->super.async_mode        = UCS_ASYNC_THREAD_LOCK_TYPE;
     params->super.wait_mode         = UCX_PERF_WAIT_MODE_LAST;
@@ -190,11 +213,16 @@ ucs_status_t init_test_params(perftest_params_t *params)
     params->super.uct.am_hdr_size   = 8;
     params->super.send_mem_type     = UCS_MEMORY_TYPE_HOST;
     params->super.recv_mem_type     = UCS_MEMORY_TYPE_HOST;
+    params->super.send_device       = default_dev;
+    params->super.recv_device       = default_dev;
+    params->super.device_level      = UCS_DEVICE_LEVEL_THREAD;
     params->super.msg_size_cnt      = 1;
     params->super.iov_stride        = 0;
     params->super.ucp.send_datatype = UCP_PERF_DATATYPE_CONTIG;
     params->super.ucp.recv_datatype = UCP_PERF_DATATYPE_CONTIG;
     params->super.ucp.am_hdr_size   = 0;
+    params->super.device_thread_count = 1;
+    params->super.device_block_count  = 1;
     params->super.ucp.is_daemon_mode  = 0;
     params->super.ucp.dmn_local_addr  = empty_addr;
     params->super.ucp.dmn_remote_addr = empty_addr;
@@ -228,9 +256,10 @@ static unsigned sock_rte_group_index(void *rte_group)
 static void sock_rte_barrier(void *rte_group, void (*progress)(void *arg),
                              void *arg)
 {
-#pragma omp barrier
-
-#pragma omp master
+#if _OPENMP
+#  pragma omp barrier
+#  pragma omp master
+#endif
   {
     sock_rte_group_t *group = rte_group;
 
@@ -251,7 +280,10 @@ static void sock_rte_barrier(void *rte_group, void (*progress)(void *arg),
         }
     }
   }
-#pragma omp barrier
+
+#if _OPENMP
+#  pragma omp barrier
+#endif
 }
 
 static void sock_rte_post_vec(void *rte_group, const struct iovec *iovec,
@@ -345,7 +377,7 @@ static ucs_status_t setup_sock_rte_p2p(struct perftest_context *ctx)
     int ret;
     char service[8];
     char err_str[64];
-    unsigned needed_flags;
+    perftest_params_t peer_params;
 
     ucs_snprintf_safe(service, sizeof(service), "%u", ctx->port);
     memset(&hints, 0, sizeof(hints));
@@ -425,38 +457,38 @@ static ucs_status_t setup_sock_rte_p2p(struct perftest_context *ctx)
     }
 
     if (ctx->server_addr == NULL) {
-        /* release the memory for the list of the message sizes allocated
-         * during the initialization of the default testing parameters */
-        release_msg_size_list(&ctx->params);
-
-        needed_flags = ctx->params.super.flags &
-                       UCX_PERF_TEST_FLAG_ERR_HANDLING;
-        ret = safe_recv(connfd, &ctx->params, sizeof(ctx->params), NULL, NULL);
+        ret = safe_recv(connfd, &peer_params, sizeof(peer_params), NULL, NULL);
         if (ret) {
             status = UCS_ERR_IO_ERROR;
             goto err_close_connfd;
         }
-        ctx->params.super.flags |= needed_flags;
 
-        if (ctx->params.super.msg_size_cnt != 0) {
-            ctx->params.super.msg_size_list =
-                    calloc(ctx->params.super.msg_size_cnt,
-                           sizeof(*ctx->params.super.msg_size_list));
-            if (NULL == ctx->params.super.msg_size_list) {
+        if (peer_params.super.msg_size_cnt != 0) {
+            peer_params.super.msg_size_list =
+                    calloc(peer_params.super.msg_size_cnt,
+                           sizeof(*peer_params.super.msg_size_list));
+            if (peer_params.super.msg_size_list == NULL) {
                 status = UCS_ERR_NO_MEMORY;
                 goto err_close_connfd;
             }
 
-            ret = safe_recv(connfd, ctx->params.super.msg_size_list,
-                            sizeof(*ctx->params.super.msg_size_list) *
-                            ctx->params.super.msg_size_cnt,
+            ret = safe_recv(connfd, peer_params.super.msg_size_list,
+                            sizeof(*peer_params.super.msg_size_list) *
+                            peer_params.super.msg_size_cnt,
                             NULL, NULL);
             if (ret) {
+                perftest_params_release_msg_size_list(&peer_params);
                 status = UCS_ERR_IO_ERROR;
                 goto err_close_connfd;
             }
         }
 
+        status = perftest_params_merge(&ctx->params, &peer_params);
+        perftest_params_release_msg_size_list(&peer_params);
+        if (status != UCS_OK) {
+            goto err_close_connfd;
+        }
+
         ctx->sock_rte_group.sendfd    = connfd;
         ctx->sock_rte_group.recvfd    = connfd;
         ctx->sock_rte_group.peer      = 1;
