@@ -36,9 +36,21 @@ enum {
     /*
      * Memory handle was imported and points to some peer's memory buffer.
      */
-    UCP_MEMH_FLAG_IMPORTED     = UCS_BIT(0),
-    UCP_MEMH_FLAG_MLOCKED      = UCS_BIT(1),
-    UCP_MEMH_FLAG_HAS_AUTO_GVA = UCS_BIT(2),
+    UCP_MEMH_FLAG_IMPORTED           = UCS_BIT(0),
+    UCP_MEMH_FLAG_MLOCKED            = UCS_BIT(1),
+    UCP_MEMH_FLAG_HAS_AUTO_GVA       = UCS_BIT(2),
+
+    /**
+     * Avoid using registration cache for the particular memory region.
+     */
+    UCP_MEMH_FLAG_NO_RCACHE          = UCS_BIT(3),
+
+    /**
+     * Track if sender-side flush is needed, check is only done when needed
+     * and cached.
+     */
+    UCP_MEMH_FLAG_SEND_FLUSH_CHECKED = UCS_BIT(4),
+    UCP_MEMH_FLAG_SEND_FLUSH_NEEDED  = UCS_BIT(5)
 };
 
 
@@ -103,6 +115,7 @@ typedef struct ucp_rndv_frag_mp_chunk_hdr {
 typedef struct ucp_rndv_mpool_priv {
     ucp_worker_h        worker;
     ucs_memory_type_t   mem_type;
+    ucs_sys_device_t    sys_dev;
 } ucp_rndv_mpool_priv_t;
 
 
@@ -112,6 +125,17 @@ typedef struct {
 } ucp_mem_dummy_handle_t;
 
 
+/**
+ * Memory type pack/unpack registration context
+ */
+typedef struct {
+    ucp_md_index_t    md_index; /* index of MD */
+    ucp_mem_h         ucp_memh; /* memh from rcache if MD is cacheable */
+    uct_mem_h         uct_memh; /* memh for specific MD */
+    uct_rkey_bundle_t rkey_bundle; /* rkey bundle from memh */
+} ucp_mtype_pack_context_t;
+
+
 extern ucp_mem_dummy_handle_t ucp_mem_dummy_handle;
 
 extern const ucp_memory_info_t ucp_mem_info_unknown;
@@ -162,11 +186,11 @@ ucs_status_t ucp_mem_rereg_mds(ucp_context_h context, ucp_md_map_t reg_md_map,
 
 ucs_status_t ucp_mem_type_reg_buffers(ucp_worker_h worker, void *remote_addr,
                                       size_t length, ucs_memory_type_t mem_type,
-                                      ucp_md_index_t md_index, ucp_mem_h memh,
-                                      uct_rkey_bundle_t *rkey_bundle);
+                                      ucp_md_index_t md_index,
+                                      ucp_mtype_pack_context_t *pack_context);
 
-void ucp_mem_type_unreg_buffers(ucp_worker_h worker, ucp_md_index_t md_index,
-                                ucp_mem_h memh, uct_rkey_bundle_t *rkey_bundle);
+void ucp_mem_type_unreg_buffers(ucp_worker_h worker,
+                                const ucp_mtype_pack_context_t *pack_context);
 
 ucs_status_t ucp_memh_get_slow(ucp_context_h context, void *address,
                                size_t length, ucs_memory_type_t mem_type,
@@ -195,19 +219,27 @@ void ucp_memh_disable_gva(ucp_mem_h memh, ucp_md_map_t md_map);
  *
  * @param [in]  context        UCP context containing memory domain indexes to
  *                             use for the memory allocation.
- * @param [in]  alloc_mem_type Memory type to get allocation index and sys
- *                             device for.
- * @param [out] md_idx         Index of the memory domain that is used to
+ * @param [in]  alloc_mem_type Memory type to get allocation index and memory
+ *                             information for.
+ * @param [in]  alloc_sys_dev  System device to get allocation index and memory
+ *                             information for.
+ * @param [out] md_idx_p       Index of the memory domain that is used to
  *                             allocate memory.
- * @param [out] sys_dev        Device id on which the memory was allocated.
+ * @param [out] mem_info_p     Information about the allocated memory.
  *
  * @return Error code as defined by @ref ucs_status_t.
  */
-ucs_status_t
-ucp_mm_get_alloc_md_index(ucp_context_h context,
-                          ucs_memory_type_t alloc_mem_type,
-                          ucp_md_index_t *md_idx,
-                          ucs_sys_device_t *sys_dev);
+ucs_status_t ucp_mm_get_alloc_md_index(ucp_context_h context,
+                                       ucs_memory_type_t alloc_mem_type,
+                                       ucs_sys_device_t alloc_sys_dev,
+                                       ucp_md_index_t *md_idx_p,
+                                       ucp_memory_info_t *mem_info_p);
+
+ucs_status_t ucp_mem_do_alloc(ucp_context_h context, void *address,
+                              size_t length, unsigned uct_flags,
+                              ucs_memory_type_t mem_type,
+                              ucs_sys_device_t sys_dev, const char *name,
+                              uct_allocated_memory_t *mem);
 
 static UCS_F_ALWAYS_INLINE ucp_md_map_t
 ucp_rkey_packed_md_map(const void *rkey_buffer)
@@ -221,16 +253,6 @@ ucp_rkey_packed_mem_type(const void *rkey_buffer)
     return (ucs_memory_type_t)(*(uint8_t *)((const ucp_md_map_t*)rkey_buffer + 1));
 }
 
-static UCS_F_ALWAYS_INLINE uct_mem_h
-ucp_memh_map2uct(const uct_mem_h *uct, ucp_md_map_t md_map, ucp_md_index_t md_idx)
-{
-    if (!(md_map & UCS_BIT(md_idx))) {
-        return UCT_MEM_HANDLE_NULL;
-    }
-
-    return uct[ucs_bitmap2idx(md_map, md_idx)];
-}
-
 static UCS_F_ALWAYS_INLINE void *ucp_memh_address(const ucp_mem_h memh)
 {
     return (void*)memh->super.super.start;
