@@ -11,6 +11,7 @@
 #include "proto_init.h"
 #include "proto_debug.h"
 #include "proto_select.inl"
+#include "proto_common.inl"
 
 #include <ucp/core/ucp_ep.inl>
 #include <ucs/datastruct/array.h>
@@ -170,7 +171,8 @@ ucp_proto_init_add_tl_perf(const ucp_proto_common_init_params_t *params,
 
     /* Send time is representing request completion, which in case of zcopy
        waits for ACK from remote side. */
-    if (params->flags & UCP_PROTO_COMMON_INIT_FLAG_SEND_ZCOPY) {
+    if ((op_attr_mask & UCP_OP_ATTR_FLAG_FAST_CMPL) &&
+        (params->flags & UCP_PROTO_COMMON_INIT_FLAG_SEND_ZCOPY)) {
         perf_factors[UCP_PROTO_PERF_FACTOR_LATENCY].c += tl_perf->latency;
     }
 
@@ -182,7 +184,59 @@ ucp_proto_init_add_tl_perf(const ucp_proto_common_init_params_t *params,
     }
 
     return ucp_proto_perf_add_funcs(perf, range_start, range_end, perf_factors,
-                                    tl_perf_node, "transport", "");
+                                    ucp_proto_perf_node_new_data("transport",
+                                                                 ""),
+                                    tl_perf_node);
+}
+
+/**
+ * This routine creates performance calculation tree node representing cost of
+ * memory registration (MR) on memory domains (MD) from @a md_map_p. It excludes
+ * MDs with zero MR cost from @a md_map_p, and adds the summarized MR cost to
+ * @a perf_factor.
+ *
+ * @param [in]     context     UCP context
+ * @param [in,out] md_map_p    Map of MDs
+ * @param [in,out] perf_factor Performance factor
+ *
+ * @return Performance node containing MR cost. NULL if total MR cost is zero.
+ */
+static ucp_proto_perf_node_t *
+ucp_proto_init_memreg_node(ucp_context_h context, ucp_md_map_t *md_map_p,
+                           ucs_linear_func_t *perf_factor)
+{
+    ucp_proto_perf_node_t *perf_node;
+    ucp_md_index_t md_index;
+    ucs_linear_func_t reg_cost;
+    const char *md_name;
+
+    perf_node = ucp_proto_perf_node_new_data("mem reg", "");
+    ucs_for_each_bit(md_index, *md_map_p) {
+        reg_cost = context->tl_mds[md_index].attr.reg_cost;
+        if (ucs_linear_func_is_zero(reg_cost, UCP_PROTO_PERF_EPSILON)) {
+            *md_map_p &= ~UCS_BIT(md_index);
+            continue;
+        }
+
+        md_name = context->tl_mds[md_index].rsc.md_name;
+        ucs_linear_func_add_inplace(perf_factor, reg_cost);
+        ucs_trace("md %s reg: " UCP_PROTO_PERF_FUNC_FMT, md_name,
+                  UCP_PROTO_PERF_FUNC_ARG(&reg_cost));
+        ucp_proto_perf_node_add_data(perf_node, md_name, reg_cost);
+    }
+
+    if (*md_map_p == 0) {
+        /* Noting to add */
+        ucp_proto_perf_node_deref(&perf_node);
+        return NULL;
+    }
+
+    if (!ucs_is_pow2(*md_map_p)) {
+        /* Multiple memory domains */
+        ucp_proto_perf_node_add_data(perf_node, "total", *perf_factor);
+    }
+
+    return perf_node;
 }
 
 ucs_status_t
@@ -194,11 +248,9 @@ ucp_proto_init_add_memreg_time(const ucp_proto_common_init_params_t *params,
 {
     ucp_context_h context                 = params->super.worker->context;
     ucp_proto_perf_factors_t perf_factors = UCP_PROTO_PERF_FACTORS_INITIALIZER;
+    ucp_proto_perf_node_t *perf_node;
     ucp_proto_perf_node_t *reg_perf_node;
-    const uct_md_attr_v2_t *md_attr;
-    ucp_md_index_t md_index;
     ucs_status_t status;
-    const char *md_name;
 
     if (reg_md_map == 0) {
         return UCS_OK;
@@ -207,34 +259,18 @@ ucp_proto_init_add_memreg_time(const ucp_proto_common_init_params_t *params,
     if (context->rcache != NULL) {
         perf_factors[cpu_factor_id] =
                 ucs_linear_func_make(context->config.ext.rcache_overhead, 0);
+        perf_node = ucp_proto_perf_node_new_data("rcache lookup", "");
         return ucp_proto_perf_add_funcs(perf, range_start, range_end,
-                                        perf_factors, NULL, "rcache lookup",
-                                        "");
+                                        perf_factors, perf_node, NULL);
     }
 
-    reg_perf_node = ucp_proto_perf_node_new_data("mem reg", "");
-
-    /* Go over all memory domains */
-    ucs_for_each_bit(md_index, reg_md_map) {
-        md_attr = &context->tl_mds[md_index].attr;
-        md_name = context->tl_mds[md_index].rsc.md_name;
-        ucs_linear_func_add_inplace(&perf_factors[cpu_factor_id],
-                                    md_attr->reg_cost);
-        ucs_trace("md %s reg: " UCP_PROTO_PERF_FUNC_FMT, md_name,
-                  UCP_PROTO_PERF_FUNC_ARG(&md_attr->reg_cost));
-        ucp_proto_perf_node_add_data(reg_perf_node, md_name, md_attr->reg_cost);
-    }
-
-    if (!ucs_is_pow2(reg_md_map)) {
-        /* Multiple memory domains */
-        ucp_proto_perf_node_add_data(reg_perf_node, "total",
-                                     perf_factors[cpu_factor_id]);
-    }
-
-    status = ucp_proto_perf_add_funcs(perf, range_start, range_end,
-                                      perf_factors, reg_perf_node,
-                                      perf_node_name, "%u mds",
-                                      ucs_popcount(reg_md_map));
+    reg_perf_node = ucp_proto_init_memreg_node(context, &reg_md_map,
+                                               &perf_factors[cpu_factor_id]);
+    perf_node     = ucp_proto_perf_node_new_data(perf_node_name, "%u mds",
+                                                 ucs_popcount(reg_md_map));
+    status        = ucp_proto_perf_add_funcs(perf, range_start, range_end,
+                                             perf_factors, perf_node,
+                                             reg_perf_node);
     ucp_proto_perf_node_deref(&reg_perf_node);
     return status;
 }
@@ -280,6 +316,9 @@ ucp_proto_init_add_buffer_copy_time(ucp_worker_h worker, const char *title,
     ucp_rsc_index_t rsc_index;
     ucp_lane_index_t lane;
     ucs_status_t status;
+    ucp_proto_perf_node_t *reg_perf_node;
+    ucp_proto_perf_node_t *perf_node;
+    ucp_md_map_t md_map;
 
     buffer_copy_factor_id = ucp_proto_buffer_copy_factor_id(local_mem_type,
                                                             remote_mem_type,
@@ -287,8 +326,9 @@ ucp_proto_init_add_buffer_copy_time(ucp_worker_h worker, const char *title,
     if (UCP_MEM_IS_HOST(local_mem_type) && UCP_MEM_IS_HOST(remote_mem_type)) {
         perf_factors[buffer_copy_factor_id] =
                 ucs_linear_func_make(0, 1.0 / context->config.ext.bcopy_bw);
+        perf_node = ucp_proto_perf_node_new_data(title, "memcpy");
         return ucp_proto_perf_add_funcs(perf, range_start, range_end,
-                                        perf_factors, NULL, title, "memcpy");
+                                        perf_factors, perf_node, NULL);
     }
 
     if (worker->mem_type_ep[local_mem_type] != NULL) {
@@ -348,7 +388,7 @@ ucp_proto_init_add_buffer_copy_time(ucp_worker_h worker, const char *title,
     perf_factors[buffer_copy_factor_id].c +=
             ucp_tl_iface_latency(context, &perf_attr.latency);
     perf_factors[buffer_copy_factor_id].m +=
-            1.0 / ucp_tl_iface_bandwidth(context, &perf_attr.bandwidth);
+            1.0 / ucp_proto_common_iface_bandwidth(context, &perf_attr.bandwidth);
 
     if ((memtype_op == UCT_EP_OP_GET_SHORT) ||
         (memtype_op == UCT_EP_OP_GET_ZCOPY)) {
@@ -359,14 +399,20 @@ ucp_proto_init_add_buffer_copy_time(ucp_worker_h worker, const char *title,
         dst_mem_type = remote_mem_type;
     }
 
+    perf_node = ucp_proto_perf_node_new_data(
+            title, "%s to %s", ucs_memory_type_names[src_mem_type],
+            ucs_memory_type_names[dst_mem_type]);
+
     ucp_proto_common_lane_perf_node(context, rsc_index, &perf_attr,
                                     &tl_perf_node);
 
+    md_map        = UCS_BIT(context->tl_rscs[rsc_index].md_index);
+    reg_perf_node = ucp_proto_init_memreg_node(
+            context, &md_map, &perf_factors[buffer_copy_factor_id]);
+
+    ucp_proto_perf_node_own_child(perf_node, &reg_perf_node);
     status = ucp_proto_perf_add_funcs(perf, range_start, range_end,
-                                      perf_factors, tl_perf_node, title,
-                                      "%s to %s",
-                                      ucs_memory_type_names[src_mem_type],
-                                      ucs_memory_type_names[dst_mem_type]);
+                                      perf_factors, perf_node, tl_perf_node);
     ucp_proto_perf_node_deref(&tl_perf_node);
 
     return status;
@@ -519,6 +565,23 @@ err_cleanup_perf:
     return status;
 }
 
+int ucp_proto_common_check_memtype_copy(
+    const ucp_proto_common_init_params_t *params)
+{
+    if (params->super.worker->context->config.ext.memtype_copy_enable) {
+        return 1;
+    }
+
+    if (params->flags & UCP_PROTO_COMMON_INIT_FLAG_HDR_ONLY) {
+        return 1;
+    }
+
+    return ucs_test_flags(params->flags,
+                          UCP_PROTO_COMMON_INIT_FLAG_SEND_ZCOPY,
+                          UCP_PROTO_COMMON_INIT_FLAG_RECV_ZCOPY) &&
+           (params->flags & UCP_PROTO_COMMON_INIT_FLAG_REMOTE_ACCESS);
+}
+
 int ucp_proto_init_check_op(const ucp_proto_init_params_t *init_params,
                             uint64_t op_id_mask)
 {
