@@ -23,7 +23,8 @@ ucp_proto_rndv_ctrl_get_md_map(const ucp_proto_rndv_ctrl_init_params_t *params,
 {
     ucp_context_h context                    = params->super.super.worker->context;
     const ucp_ep_config_key_t *ep_config_key = params->super.super.ep_config_key;
-    ucp_rsc_index_t mem_sys_dev, ep_sys_dev;
+    ucp_rsc_index_t mem_sys_dev              = params->super.reg_mem_info.sys_dev;
+    ucp_rsc_index_t ep_sys_dev;
     const uct_iface_attr_t *iface_attr;
     const uct_md_attr_v2_t *md_attr;
     const uct_component_attr_t *cmpt_attr;
@@ -31,8 +32,8 @@ ucp_proto_rndv_ctrl_get_md_map(const ucp_proto_rndv_ctrl_init_params_t *params,
     ucp_lane_index_t lane;
     ucs_status_t status;
 
-    /* md_map is all lanes which support get_zcopy on the given mem_type and
-     * require remote key
+    /* md_map is all lanes which support get_zcopy on the given mem_type,
+     * require remote key and can reach memory sys_dev
      */
     *md_map      = 0;
     *sys_dev_map = 0;
@@ -76,6 +77,11 @@ ucp_proto_rndv_ctrl_get_md_map(const ucp_proto_rndv_ctrl_init_params_t *params,
             continue;
         }
 
+        /* Check reachability between mem_sys_dev and current lane's sys_dev */
+        if (!ucs_topo_is_reachable(ep_sys_dev, mem_sys_dev)) {
+            continue;
+        }
+
         ucs_trace_req("lane[%d]: selected md %s index %u", lane,
                       context->tl_mds[md_index].rsc.md_name, md_index);
         *md_map |= UCS_BIT(md_index);
@@ -87,7 +93,6 @@ ucp_proto_rndv_ctrl_get_md_map(const ucp_proto_rndv_ctrl_init_params_t *params,
         *sys_dev_map |= UCS_BIT(ep_sys_dev);
     }
 
-    mem_sys_dev = params->super.reg_mem_info.sys_dev;
     ucs_for_each_bit(ep_sys_dev, *sys_dev_map) {
         status = ucs_topo_get_distance(mem_sys_dev, ep_sys_dev, sys_distance);
         ucs_assertv_always(status == UCS_OK, "mem_info->sys_dev=%d sys_dev=%d",
@@ -153,6 +158,7 @@ static ucs_status_t ucp_proto_rndv_ctrl_select_remote_proto(
     rkey_config_key.ep_cfg_index = ep_cfg_index;
     rkey_config_key.sys_dev      = params->super.reg_mem_info.sys_dev;
     rkey_config_key.mem_type     = params->super.reg_mem_info.type;
+    rkey_config_key.flags        = 0;
 
     rkey_config_key.unreachable_md_map = 0;
 
@@ -200,6 +206,7 @@ ucp_proto_rndv_ctrl_perf(const ucp_proto_common_init_params_t *init_params,
     ucp_rsc_index_t rsc_index;
     ucp_proto_perf_t *perf;
     ucs_status_t status;
+    ucp_proto_perf_node_t *perf_node;
 
     if (lane == UCP_NULL_LANE) {
         return UCS_ERR_NO_ELEM;
@@ -237,9 +244,10 @@ ucp_proto_rndv_ctrl_perf(const ucp_proto_common_init_params_t *init_params,
     perf_factors[UCP_PROTO_PERF_FACTOR_LATENCY]    = ucs_linear_func_make(
             ucp_tl_iface_latency(worker->context, &perf_attr.latency), 0.0);
 
-    status = ucp_proto_perf_add_funcs(perf, init_params->min_length,
-                                      init_params->max_length, perf_factors,
-                                      NULL, name, "");
+    perf_node = ucp_proto_perf_node_new_data(name, "");
+    status    = ucp_proto_perf_add_funcs(perf, init_params->min_length,
+                                         init_params->max_length, perf_factors,
+                                         perf_node, NULL);
     if (status != UCS_OK) {
         goto err_destroy_perf;
     }
@@ -306,7 +314,7 @@ ucp_proto_rndv_ctrl_init_priv(const ucp_proto_rndv_ctrl_init_params_t *params,
     rpriv->lane             = lane;
     rpriv->packed_rkey_size = ucp_rkey_packed_size(
             init_params->worker->context, rpriv->md_map,
-            params->super.reg_mem_info.sys_dev, rpriv->sys_dev_map);
+            params->super.reg_mem_info.sys_dev, rpriv->sys_dev_map, 0);
 }
 
 void ucp_proto_rndv_set_variant_config(
@@ -395,24 +403,21 @@ static void ucp_proto_rndv_ctrl_variant_probe(
     /* Set priority and threshold for this variant */
     cfg_thresh   = params->super.cfg_thresh;
     cfg_priority = params->super.cfg_priority;
-    if ((remote_proto->cfg_thresh != UCS_MEMUNITS_INF) &&
-        (remote_proto->cfg_thresh != UCS_MEMUNITS_AUTO)) {
-        /* Consider remote priority and threshold only if RNDV_SCHEME or
-            * BCOPY/ZCOPY thresh are set to force these settings */
-        cfg_priority = remote_proto->cfg_priority;
-        cfg_thresh   = (cfg_thresh == UCS_MEMUNITS_AUTO) ?
-                                 remote_proto->cfg_thresh :
-                                 ucs_max(cfg_thresh, remote_proto->cfg_thresh);
+    if (remote_proto->cfg_thresh != UCS_MEMUNITS_AUTO) {
+        /* If RNDV_SCHEME is set, all protocols except forced one report INF */
+        ucs_assertv(remote_proto->cfg_thresh == UCS_MEMUNITS_INF,
+                    "variant_name=%s remote_proto->cfg_thresh=%zu",
+                    variant_name, remote_proto->cfg_thresh);
+        cfg_thresh = remote_proto->cfg_thresh;
+    }
+
+    if (fabs(params->perf_bias) > UCP_PROTO_PERF_EPSILON) {
+        ucp_proto_perf_apply_func(perf,
+                                  ucs_linear_func_make(0.0,
+                                                       1.0 - params->perf_bias),
+                                  "bias", "%.2f %%", params->perf_bias);
     }
 
-    /* Remote variants priorities are used to respect RNDV_SCHEME setting
-     * so they should contain value greater than CTRL message `cfg_thresh`.
-     * Equality is allowed for RTR remote variants.
-     */
-    ucs_assertv(params->super.cfg_priority <= remote_proto->cfg_priority,
-                "remote_proto=%s params->super.cfg_priority=%u "
-                "remote_proto->cfg_priority=%u", variant_name,
-                params->super.cfg_priority, remote_proto->cfg_priority);
     ucp_proto_select_add_proto(&params->super.super, cfg_thresh, cfg_priority,
                                perf, rpriv, priv_size);
 
@@ -503,12 +508,10 @@ ucp_proto_rndv_find_ctrl_lane(const ucp_proto_init_params_t *params)
 {
     ucp_lane_index_t lane, num_lanes;
 
-    num_lanes = ucp_proto_common_find_lanes(params, UCT_EP_OP_LAST,
+    num_lanes = ucp_proto_common_find_lanes(params,
                                             UCP_PROTO_COMMON_INIT_FLAG_HDR_ONLY,
-                                            UCP_PROTO_COMMON_OFFSET_INVALID, 1,
                                             UCP_LANE_TYPE_AM,
-                                            UCS_MEMORY_TYPE_UNKNOWN,
-                                            UCT_IFACE_FLAG_AM_BCOPY, 1, 0,
+                                            UCT_IFACE_FLAG_AM_BCOPY, 1, 0, NULL,
                                             &lane);
     if (num_lanes == 0) {
         ucs_debug("no active message lane for %s",
@@ -646,6 +649,7 @@ ucp_proto_rndv_bulk_init(const ucp_proto_multi_init_params_t *init_params,
     }
 
     rpriv->frag_mem_type = init_params->super.reg_mem_info.type;
+    rpriv->frag_sys_dev  = init_params->super.reg_mem_info.sys_dev;
 
     if (rpriv->super.lane == UCP_NULL_LANE) {
         /* Add perf without ACK in case of pipeline */
@@ -673,6 +677,10 @@ size_t ucp_proto_rndv_common_pack_ack(void *dest, void *arg)
 ucs_status_t ucp_proto_rndv_ats_complete(ucp_request_t *req)
 {
     ucp_datatype_iter_cleanup(&req->send.state.dt_iter, 1, UCP_DT_MASK_ALL);
+    if (req->send.rndv.rkey != NULL) {
+        ucp_proto_rndv_rkey_destroy(req);
+    }
+
     return ucp_proto_rndv_recv_complete(req);
 }
 
@@ -716,6 +724,7 @@ UCS_PROFILE_FUNC(ucs_status_t, ucp_proto_rndv_send_reply,
     ucp_worker_cfg_index_t rkey_cfg_index;
     ucp_proto_select_param_t sel_param;
     ucp_proto_select_t *proto_select;
+    ucs_sys_device_t sys_dev;
     ucs_status_t status;
     ucp_rkey_h rkey;
 
@@ -723,6 +732,11 @@ UCS_PROFILE_FUNC(ucs_status_t, ucp_proto_rndv_send_reply,
                (op_id < UCP_OP_ID_RNDV_LAST));
 
     if (rkey_length > 0) {
+        sys_dev = req->send.state.dt_iter.mem_info.sys_dev;
+        if (sys_dev == UCS_SYS_DEVICE_ID_UNKNOWN) {
+            sys_dev = worker->context->alloc_md[UCS_MEMORY_TYPE_CUDA].sys_dev;
+        }
+
         ucs_assert(rkey_buffer != NULL);
         /* Do not unpack rkeys from MDs with rkey_ptr capability, except
          * rkey_ptr_lane. Examples are: sysv and posix. Such keys, if packed,
@@ -730,8 +744,8 @@ UCS_PROFILE_FUNC(ucs_status_t, ucp_proto_rndv_send_reply,
          * done by the specific protocols (if selected) which use them.
          */
         status = ucp_ep_rkey_unpack_internal(
-                  ep, rkey_buffer, rkey_length, ep_config->key.reachable_md_map,
-                  ep_config->rndv.proto_rndv_rkey_skip_mds, &rkey);
+                ep, rkey_buffer, rkey_length, ep_config->key.reachable_md_map,
+                ep_config->rndv.proto_rndv_rkey_skip_mds, sys_dev, &rkey);
         if (status != UCS_OK) {
             goto err;
         }
