@@ -67,6 +67,9 @@ typedef cpuset_t ucs_sys_cpuset_t;
 #define UCS_SYS_FS_CPUS_PATH   UCS_SYS_FS_SYSTEM_PATH "/cpu"
 
 
+#define UCS_SYS_DMI_PRODUCT_NAME_LABEL "Product name"
+
+
 BEGIN_C_DECLS
 
 /** @file sys.h */
@@ -183,6 +186,22 @@ const char *ucs_get_exe();
 uint32_t ucs_file_checksum(const char *filename);
 
 
+/**
+ * Get interface index for a given interface name.
+ */
+ucs_status_t ucs_ifname_to_index(const char *ndev_name, unsigned *ndev_index_p);
+
+
+/**
+ * Check if a network interface is an IPoIB (IP over InfiniBand) device.
+ *
+ * @param [in]  if_name  Network interface name to check.
+ *
+ * @return 1 if the interface is IPoIB, 0 otherwise.
+ */
+int ucs_netif_is_ipoib(const char *if_name);
+
+
 /**
  * Get a globally unique identifier of the machine running the current process.
  */
@@ -226,7 +245,7 @@ int ucs_sys_max_open_files();
  * @param [in]  err_log_level  Logging level that should be used for printing
  *                             errors.
  * @param [out] p_fstream      Pointer that is filled with the stream handle.
- *                             User is responsible to close tha stream handle then.
+ *                             User is responsible to close the stream handle then.
  * @param [out] p_need_close   Pointer to the variable that is set to whether
  *                             fclose() should be called to release resources (1)
  *                             or not (0).
@@ -690,6 +709,12 @@ ucs_status_t ucs_sys_get_effective_memlock_rlimit(size_t *rlimit_value);
  */
 int ucs_sys_is_dynamic_lib(void);
 
+
+/**
+ * @return Product name from DMI table.
+ */
+const char *ucs_sys_dmi_product_name();
+
 END_C_DECLS
 
 #endif
