@@ -17,10 +17,20 @@
 #define UCT_GGA_MLX5_OPAQUE_BUF_LEN 64
 #define UCT_GGA_MAX_MSG_SIZE        (2u * UCS_MBYTE)
 
+#if ENABLE_ASSERT
+#define UCT_GGA_MLX5_MD_CAPS        (UCT_IB_MLX5_MD_FLAG_DEVX | \
+                                     UCT_IB_MLX5_MD_FLAG_MMO_DMA)
+#endif /* ENABLE_ASSERT */
+
+typedef struct {
+    uct_ib_mlx5_md_t super;
+    pthread_mutex_t  mem_attach_lock;
+} uct_gga_mlx5_md_t;
+
 typedef struct {
     uct_ib_md_packed_mkey_t packed_mkey;
     uct_ib_mlx5_devx_mem_t  *memh;
-    uct_ib_mlx5_md_t        *md;
+    uct_gga_mlx5_md_t       *md;
     uct_rkey_bundle_t       rkey_ob;
 } uct_gga_mlx5_rkey_handle_t;
 
@@ -49,9 +59,9 @@ enum {
 };
 
 typedef struct {
-    uint8_t         flags;
-    uct_ib_uint24_t qp_num;
-    uct_ib_uint24_t flush_rkey;
+    uct_rc_mlx5_base_ep_address_t super;
+    uint8_t                       flags;
+    uct_ib_uint24_t               flush_rkey;
 } UCS_S_PACKED uct_gga_mlx5_ep_address_t;
 
 typedef struct {
@@ -72,6 +82,35 @@ ucs_config_field_t uct_gga_mlx5_iface_config_table[] = {
   {NULL}
 };
 
+static ucs_status_t
+uct_ib_mlx5_gga_md_query(uct_md_h uct_md, uct_md_attr_v2_t *md_attr)
+{
+    ucs_status_t status;
+
+    status = uct_ib_mlx5_devx_md_query(uct_md, md_attr);
+    if (status != UCS_OK) {
+        return status;
+    }
+
+    /* GGA always packs exported mkey */
+    md_attr->rkey_packed_size = md_attr->exported_mkey_packed_size;
+    return UCS_OK;
+}
+
+static ucs_status_t
+uct_ib_mlx5_gga_mem_attach(uct_md_h uct_md, const void *mkey_buffer,
+                           uct_md_mem_attach_params_t *params,
+                           uct_mem_h *memh_p)
+{
+    uct_gga_mlx5_md_t *gga_md = ucs_derived_of(uct_md, uct_gga_mlx5_md_t);
+    ucs_status_t status;
+
+    pthread_mutex_lock(&gga_md->mem_attach_lock);
+    status = uct_ib_mlx5_devx_mem_attach(uct_md, mkey_buffer, params, memh_p);
+    pthread_mutex_unlock(&gga_md->mem_attach_lock);
+    return status;
+}
+
 static ucs_status_t
 uct_ib_mlx5_gga_mkey_pack(uct_md_h uct_md, uct_mem_h uct_memh,
                           void *address, size_t length,
@@ -93,6 +132,7 @@ uct_ib_mlx5_gga_mkey_pack(uct_md_h uct_md, uct_mem_h uct_memh,
 
 static ucs_status_t
 uct_gga_mlx5_rkey_unpack(uct_component_t *component, const void *rkey_buffer,
+                         const uct_rkey_unpack_params_t *params,
                          uct_rkey_t *rkey_p, void **handle_p)
 {
     const uct_ib_md_packed_mkey_t *mkey = rkey_buffer;
@@ -130,7 +170,8 @@ uct_gga_mlx5_rkey_handle_dereg(uct_gga_mlx5_rkey_handle_t *rkey_handle)
         return;
     }
 
-    status = uct_ib_mlx5_devx_mem_dereg(&rkey_handle->md->super.super, &params);
+    status = uct_ib_mlx5_devx_mem_dereg(&rkey_handle->md->super.super.super,
+                                        &params);
     if (status != UCS_OK) {
         ucs_warn("md %p: failed to deregister GGA memh %p", rkey_handle->md,
                  rkey_handle->memh);
@@ -150,17 +191,71 @@ static ucs_status_t uct_gga_mlx5_rkey_release(uct_component_t *component,
     return UCS_OK;
 }
 
+static int uct_ib_mlx5_gga_check_device(struct ibv_device *device)
+{
+    int result = 0;
+    char out[UCT_IB_MLX5DV_ST_SZ_BYTES(query_hca_cap_out)];
+    struct ibv_context *ctx;
+    ucs_status_t status;
+    size_t out_len;
+    char *cap;
+
+    ctx = uct_ib_mlx5_devx_open_device(device);
+    if (ctx == NULL) {
+        goto out;
+    }
+
+    out_len = ucs_static_array_size(out);
+    status  = uct_ib_mlx5_devx_query_cap(ctx,
+                                         UCT_IB_MLX5_HCA_CAP_OPMOD_GET_CUR |
+                                         (UCT_IB_MLX5_CAP_GENERAL << 1),
+                                         out, out_len, "QUERY_HCA_CAP", 0);
+    if (status != UCS_OK) {
+        goto out_close_ctx;
+    }
+
+    cap = UCT_IB_MLX5DV_ADDR_OF(query_hca_cap_out, out, capability);
+    if (!UCT_IB_MLX5DV_GET(cmd_hca_cap, cap, dma_mmo_qp)) {
+        goto out_close_ctx;
+    }
+
+    status = uct_ib_mlx5_devx_query_cap_2(ctx, out, out_len);
+    if (status != UCS_OK) {
+        goto out_close_ctx;
+    }
+
+    /*
+     * This check does not guarantee that XGVMI works with indirect mkeys.
+     * We can filter out devices that do not support XGVMI at all.
+     */
+    result = uct_ib_mlx5_devx_check_xgvmi(cap);
+
+out_close_ctx:
+    uct_ib_md_device_context_close(ctx);
+out:
+    return result;
+}
+
+static ucs_status_t
+uct_ib_mlx5_gga_query_md_resources(uct_component_t *component,
+                                   uct_md_resource_desc_t **resources_p,
+                                   unsigned *num_resources_p)
+{
+    return uct_ib_base_query_md_resources(resources_p, num_resources_p,
+                                          uct_ib_mlx5_gga_check_device);
+}
+
 /* Forward declaration */
 static ucs_status_t
 uct_ib_mlx5_gga_md_open(uct_component_t *component, const char *md_name,
                         const uct_md_config_t *uct_md_config, uct_md_h *md_p);
 
 static uct_component_t uct_gga_component = {
-    .query_md_resources = uct_ib_query_md_resources,
+    .query_md_resources = uct_ib_mlx5_gga_query_md_resources,
     .md_open            = uct_ib_mlx5_gga_md_open,
-    .cm_open            = ucs_empty_function_return_unsupported,
+    .cm_open            = (uct_component_cm_open_func_t)ucs_empty_function_return_unsupported,
     .rkey_unpack        = uct_gga_mlx5_rkey_unpack,
-    .rkey_ptr           = ucs_empty_function_return_unsupported,
+    .rkey_ptr           = (uct_component_rkey_ptr_func_t)ucs_empty_function_return_unsupported,
     .rkey_release       = uct_gga_mlx5_rkey_release,
     .rkey_compare       = uct_base_rkey_compare,
     .name               = "gga",
@@ -176,10 +271,10 @@ static uct_component_t uct_gga_component = {
     .md_vfs_init        = (uct_component_md_vfs_init_func_t)ucs_empty_function
 };
 
-static UCS_F_ALWAYS_INLINE
-void uct_gga_mlx5_rkey_trace(uct_ib_mlx5_md_t *md,
-                             uct_gga_mlx5_rkey_handle_t *rkey_handle,
-                             const char *prefix)
+static UCS_F_ALWAYS_INLINE void
+ uct_gga_mlx5_rkey_trace(uct_gga_mlx5_md_t *md,
+                         uct_gga_mlx5_rkey_handle_t *rkey_handle,
+                         const char *prefix)
 {
     ucs_trace("md %p: %s resolved rkey %p: rkey_ob %"PRIx64"/%p", md, prefix,
               rkey_handle, rkey_handle->rkey_ob.rkey,
@@ -187,12 +282,13 @@ void uct_gga_mlx5_rkey_trace(uct_ib_mlx5_md_t *md,
 }
 
 static UCS_F_ALWAYS_INLINE ucs_status_t
-uct_gga_mlx5_rkey_resolve(uct_ib_mlx5_md_t *md,
+uct_gga_mlx5_rkey_resolve(uct_gga_mlx5_md_t *md,
                           uct_gga_mlx5_rkey_handle_t *rkey_handle)
 {
-    uct_md_h uct_md                         = &md->super.super;
-    uct_md_mem_attach_params_t atach_params = { 0 };
-    uct_md_mkey_pack_params_t repack_params = { 0 };
+    uct_md_h uct_md                          = &md->super.super.super;
+    uct_md_mem_attach_params_t attach_params = { 0 };
+    uct_md_mkey_pack_params_t repack_params  = { 0 };
+    uct_rkey_unpack_params_t unpack_params   = { 0 };
     uint64_t repack_mkey;
     ucs_status_t status;
 
@@ -201,9 +297,9 @@ uct_gga_mlx5_rkey_resolve(uct_ib_mlx5_md_t *md,
         return UCS_OK;
     }
 
-    status = uct_ib_mlx5_devx_mem_attach(uct_md, &rkey_handle->packed_mkey,
-                                         &atach_params,
-                                         (uct_mem_h *)&rkey_handle->memh);
+    status = uct_ib_mlx5_gga_mem_attach(uct_md, &rkey_handle->packed_mkey,
+                                        &attach_params,
+                                        (uct_mem_h *)&rkey_handle->memh);
     if (status != UCS_OK) {
         goto err_out;
     }
@@ -216,7 +312,7 @@ uct_gga_mlx5_rkey_resolve(uct_ib_mlx5_md_t *md,
         goto err_dereg;
     }
 
-    status = uct_ib_rkey_unpack(NULL, &repack_mkey,
+    status = uct_ib_rkey_unpack(NULL, &repack_mkey, &unpack_params,
                                 &rkey_handle->rkey_ob.rkey,
                                 &rkey_handle->rkey_ob.handle);
     uct_gga_mlx5_rkey_trace(md, rkey_handle, "new");
@@ -242,6 +338,7 @@ static ucs_status_t
 uct_gga_mlx5_iface_query(uct_iface_h tl_iface, uct_iface_attr_t *iface_attr)
 {
     uct_rc_iface_t *iface = ucs_derived_of(tl_iface, uct_rc_iface_t);
+    size_t iface_mtu      = uct_ib_mtu_value(iface->super.config.path_mtu);
     ucs_status_t status;
 
     status = uct_ib_iface_query(&iface->super, UCT_IB_RETH_LEN, iface_attr);
@@ -265,11 +362,13 @@ uct_gga_mlx5_iface_query(uct_iface_h tl_iface, uct_iface_attr_t *iface_attr)
     iface_attr->cap.put.max_zcopy       = UCT_GGA_MAX_MSG_SIZE;
     iface_attr->cap.put.max_iov         = 1;
     iface_attr->cap.put.opt_zcopy_align = UCS_SYS_PCI_MAX_PAYLOAD;
+    iface_attr->cap.put.align_mtu       = iface_mtu;
 
     iface_attr->cap.get.min_zcopy       = 1;
     iface_attr->cap.get.max_zcopy       = iface->config.max_get_zcopy;
     iface_attr->cap.get.max_iov         = 1;
     iface_attr->cap.get.opt_zcopy_align = UCS_SYS_PCI_MAX_PAYLOAD;
+    iface_attr->cap.get.align_mtu       = iface_mtu;
 
     iface_attr->latency.c += 200e-9;
 
@@ -385,7 +484,7 @@ uct_gga_mlx5_ep_get_address(uct_ep_h tl_ep, uct_ep_addr_t *addr)
     uct_gga_mlx5_ep_address_t *gga_addr = (uct_gga_mlx5_ep_address_t*)addr;
     uct_ib_md_t *md = uct_ib_iface_md(&iface->super.super);
 
-    uct_ib_pack_uint24(gga_addr->qp_num, ep->super.tx.wq.super.qp_num);
+    uct_ib_pack_uint24(gga_addr->super.qp_num, ep->super.tx.wq.super.qp_num);
     if (uct_rc_iface_flush_rkey_enabled(&iface->super)) {
         gga_addr->flags = UCT_GGA_MLX5_EP_ADDRESS_FLAG_FLUSH_RKEY;
         uct_ib_pack_uint24(gga_addr->flush_rkey, md->flush_rkey >> 8);
@@ -414,12 +513,15 @@ uct_gga_mlx5_ep_connect_to_ep_v2(uct_ep_h tl_ep,
     enum ibv_mtu path_mtu;
     ucs_status_t status;
 
-    uct_ib_iface_fill_ah_attr_from_addr(&iface->super.super, ib_addr,
-                                        ep->super.super.path_index, &ah_attr,
-                                        &path_mtu);
-    ucs_assert(path_mtu != UCT_IB_ADDRESS_INVALID_PATH_MTU);
+    status = uct_ib_iface_fill_ah_attr_from_addr(&iface->super.super, ib_addr,
+                                                 ep->super.super.path_index,
+                                                 &ah_attr, &path_mtu);
+    if (status != UCS_OK) {
+        return status;
+    }
 
-    qp_num = uct_ib_unpack_uint24(gga_ep_addr->qp_num);
+    ucs_assert(path_mtu != UCT_IB_ADDRESS_INVALID_PATH_MTU);
+    qp_num = uct_ib_unpack_uint24(gga_ep_addr->super.qp_num);
     status = uct_rc_mlx5_iface_common_devx_connect_qp(
             iface, &ep->super.tx.wq.super, qp_num, &ah_attr, path_mtu,
             ep->super.super.path_index, iface->super.config.max_rd_atomic);
@@ -456,8 +558,8 @@ uct_gga_mlx5_ep_put_zcopy(uct_ep_h tl_ep, const uct_iov_t *iov, size_t iovcnt,
 {
     UCT_RC_MLX5_BASE_EP_DECL(tl_ep, iface, ep);
     uct_gga_mlx5_ep_t *gga_ep = ucs_derived_of(ep, uct_gga_mlx5_ep_t);
-    uct_ib_mlx5_md_t *md      = ucs_derived_of(iface->super.super.super.md,
-                                               uct_ib_mlx5_md_t);
+    uct_gga_mlx5_md_t *md     = ucs_derived_of(iface->super.super.super.md,
+                                               uct_gga_mlx5_md_t);
     uct_gga_mlx5_rkey_handle_t *rkey_handle = (uct_gga_mlx5_rkey_handle_t*)rkey;
     uint8_t fm_ce_se                        = MLX5_WQE_CTRL_CQ_UPDATE;
     uct_rkey_t rkey_copy;
@@ -504,8 +606,8 @@ uct_gga_mlx5_ep_get_zcopy(uct_ep_h tl_ep, const uct_iov_t *iov, size_t iovcnt,
 {
     UCT_RC_MLX5_BASE_EP_DECL(tl_ep, iface, ep);
     uct_gga_mlx5_ep_t *gga_ep = ucs_derived_of(ep, uct_gga_mlx5_ep_t);
-    uct_ib_mlx5_md_t *md      = ucs_derived_of(iface->super.super.super.md,
-                                               uct_ib_mlx5_md_t);
+    uct_gga_mlx5_md_t *md     = ucs_derived_of(iface->super.super.super.md,
+                                               uct_gga_mlx5_md_t);
     uct_gga_mlx5_rkey_handle_t *rkey_handle = (uct_gga_mlx5_rkey_handle_t*)rkey;
     size_t total_length                     = uct_iov_total_length(iov, iovcnt);
     uint8_t fm_ce_se                        = MLX5_WQE_CTRL_CQ_UPDATE;
@@ -552,26 +654,26 @@ uct_gga_mlx5_iface_get_address(uct_iface_h tl_iface, uct_iface_addr_t *addr)
 }
 
 static uct_iface_ops_t uct_gga_mlx5_iface_tl_ops = {
-    .ep_put_short             = ucs_empty_function_return_unsupported,
+    .ep_put_short             = (uct_ep_put_short_func_t)ucs_empty_function_return_unsupported,
     .ep_put_bcopy             = (uct_ep_put_bcopy_func_t)ucs_empty_function_return_unsupported,
     .ep_put_zcopy             = uct_gga_mlx5_ep_put_zcopy,
-    .ep_get_bcopy             = ucs_empty_function_return_unsupported,
+    .ep_get_bcopy             = (uct_ep_get_bcopy_func_t)ucs_empty_function_return_unsupported,
     .ep_get_zcopy             = uct_gga_mlx5_ep_get_zcopy,
     .ep_am_short              = (uct_ep_am_short_func_t)ucs_empty_function_return_unsupported,
     .ep_am_short_iov          = (uct_ep_am_short_iov_func_t)ucs_empty_function_return_unsupported,
     .ep_am_bcopy              = (uct_ep_am_bcopy_func_t)ucs_empty_function_return_unsupported,
     .ep_am_zcopy              = (uct_ep_am_zcopy_func_t)ucs_empty_function_return_unsupported,
-    .ep_atomic_cswap64        = ucs_empty_function_return_unsupported,
-    .ep_atomic_cswap32        = ucs_empty_function_return_unsupported,
-    .ep_atomic64_post         = ucs_empty_function_return_unsupported,
-    .ep_atomic32_post         = ucs_empty_function_return_unsupported,
-    .ep_atomic64_fetch        = ucs_empty_function_return_unsupported,
-    .ep_atomic32_fetch        = ucs_empty_function_return_unsupported,
+    .ep_atomic_cswap64        = (uct_ep_atomic_cswap64_func_t)ucs_empty_function_return_unsupported,
+    .ep_atomic_cswap32        = (uct_ep_atomic_cswap32_func_t)ucs_empty_function_return_unsupported,
+    .ep_atomic64_post         = (uct_ep_atomic64_post_func_t)ucs_empty_function_return_unsupported,
+    .ep_atomic32_post         = (uct_ep_atomic32_post_func_t)ucs_empty_function_return_unsupported,
+    .ep_atomic64_fetch        = (uct_ep_atomic64_fetch_func_t)ucs_empty_function_return_unsupported,
+    .ep_atomic32_fetch        = (uct_ep_atomic32_fetch_func_t)ucs_empty_function_return_unsupported,
     .ep_pending_add           = uct_rc_ep_pending_add,
     .ep_pending_purge         = uct_rc_ep_pending_purge,
     .ep_flush                 = uct_rc_mlx5_base_ep_flush,
     .ep_fence                 = uct_rc_mlx5_base_ep_fence,
-    .ep_check                 = ucs_empty_function_return_unsupported,
+    .ep_check                 = (uct_ep_check_func_t)ucs_empty_function_return_unsupported,
     .ep_create                = UCS_CLASS_NEW_FUNC_NAME(uct_gga_mlx5_ep_t),
     .ep_destroy               = UCS_CLASS_DELETE_FUNC_NAME(uct_gga_mlx5_ep_t),
     .ep_get_address           = uct_gga_mlx5_ep_get_address,
@@ -590,6 +692,17 @@ static uct_iface_ops_t uct_gga_mlx5_iface_tl_ops = {
     .iface_is_reachable       = uct_base_iface_is_reachable
 };
 
+static int
+uct_gga_mlx5_iface_is_same_device(const uct_iface_h tl_iface,
+                                  const uct_gga_mlx5_iface_addr_t *iface_addr)
+{
+    uct_ib_iface_t *iface   = ucs_derived_of(tl_iface, uct_ib_iface_t);
+    uct_ib_device_t *device = uct_ib_iface_device(iface);
+
+    return iface_addr->be_sys_image_guid ==
+           device->dev_attr.orig_attr.sys_image_guid;
+}
+
 static int
 uct_gga_mlx5_iface_is_reachable_v2(const uct_iface_h tl_iface,
                                    const uct_iface_is_reachable_params_t *params)
@@ -601,33 +714,61 @@ uct_gga_mlx5_iface_is_reachable_v2(const uct_iface_h tl_iface,
             UCS_PARAM_VALUE(UCT_IFACE_IS_REACHABLE_FIELD, params, iface_addr,
                             IFACE_ADDR, NULL);
 
+    if (iface_addr == NULL) {
+        uct_iface_fill_info_str_buf(params, "iface address is empty");
+        return 0;
+    }
+
+    if (!uct_gga_mlx5_iface_is_same_device(tl_iface, iface_addr)) {
+        uct_iface_fill_info_str_buf(
+                params,
+                "different GUID 0x%"PRIx64" (local) vs 0x%"PRIx64" (remote)",
+                be64toh(device->dev_attr.orig_attr.sys_image_guid),
+                be64toh(iface_addr->be_sys_image_guid));
+        return 0;
+    }
+
+    return uct_ib_iface_is_reachable_v2(tl_iface, params);
+}
+
+static int
+uct_gga_mlx5_ep_is_connected(uct_ep_h tl_ep,
+                             const uct_ep_is_connected_params_t *params)
+{
+    const uct_gga_mlx5_iface_addr_t *iface_addr =
+            (const uct_gga_mlx5_iface_addr_t*)
+            UCS_PARAM_VALUE(UCT_EP_IS_CONNECTED_FIELD, params, iface_addr,
+                            IFACE_ADDR, NULL);
+
     return (iface_addr != NULL) &&
-           (iface_addr->be_sys_image_guid ==
-            device->dev_attr.orig_attr.sys_image_guid) &&
-           uct_ib_iface_is_reachable_v2(tl_iface, params);
+           uct_gga_mlx5_iface_is_same_device(tl_ep->iface, iface_addr) &&
+           uct_rc_mlx5_base_ep_is_connected(tl_ep, params);
 }
 
 static uct_rc_iface_ops_t uct_gga_mlx5_iface_ops = {
     .super = {
         .super = {
-            .iface_estimate_perf   = uct_rc_iface_estimate_perf,
-            .iface_vfs_refresh     = uct_rc_iface_vfs_refresh,
-            .ep_query              = (uct_ep_query_func_t)ucs_empty_function,
-            .ep_invalidate         = uct_rc_mlx5_base_ep_invalidate,
-            .ep_connect_to_ep_v2   = uct_gga_mlx5_ep_connect_to_ep_v2,
-            .iface_is_reachable_v2 = uct_gga_mlx5_iface_is_reachable_v2,
-            .ep_is_connected       = ucs_empty_function_do_assert
+            .iface_query_v2         = uct_iface_base_query_v2,
+            .iface_estimate_perf    = uct_rc_iface_estimate_perf,
+            .iface_vfs_refresh      = uct_rc_iface_vfs_refresh,
+            .iface_mem_element_pack = (uct_iface_mem_element_pack_func_t)ucs_empty_function_return_unsupported,
+            .ep_query               = (uct_ep_query_func_t)ucs_empty_function,
+            .ep_invalidate          = uct_rc_mlx5_base_ep_invalidate,
+            .ep_connect_to_ep_v2    = uct_gga_mlx5_ep_connect_to_ep_v2,
+            .iface_is_reachable_v2  = uct_gga_mlx5_iface_is_reachable_v2,
+            .ep_is_connected        = uct_gga_mlx5_ep_is_connected,
+            .ep_get_device_ep       = (uct_ep_get_device_ep_func_t)ucs_empty_function_return_unsupported
         },
         .create_cq      = uct_rc_mlx5_iface_common_create_cq,
         .destroy_cq     = uct_rc_mlx5_iface_common_destroy_cq,
         .event_cq       = uct_rc_mlx5_iface_common_event_cq,
         .handle_failure = uct_rc_mlx5_iface_handle_failure,
     },
-    .init_rx         = ucs_empty_function_return_success,
-    .cleanup_rx      = ucs_empty_function,
-    .fc_ctrl         = ucs_empty_function_return_unsupported,
+    .init_rx         = (uct_rc_iface_init_rx_func_t)ucs_empty_function_return_success,
+    .cleanup_rx      = (uct_rc_iface_cleanup_rx_func_t)ucs_empty_function,
+    .fc_ctrl         = (uct_rc_iface_fc_ctrl_func_t)ucs_empty_function_return_unsupported,
     .fc_handler      = (uct_rc_iface_fc_handler_func_t)ucs_empty_function_do_assert,
-    .cleanup_qp      = ucs_empty_function_do_assert_void,
+    .cleanup_qp      = (uct_rc_iface_qp_cleanup_func_t)ucs_empty_function_do_assert_void,
     .ep_post_check   = uct_rc_mlx5_base_ep_post_check,
     .ep_vfs_populate = uct_rc_mlx5_base_ep_vfs_populate
 };
@@ -649,12 +790,22 @@ static UCS_CLASS_INIT_FUNC(uct_gga_mlx5_iface_t,
     uct_ib_mlx5_md_t *md                = ucs_derived_of(tl_md, uct_ib_mlx5_md_t);
     uct_ib_iface_init_attr_t init_attr  = {};
     ucs_status_t status;
+    uct_ib_mlx5_dp_ordering_t dp_ordering;
 
     init_attr.qp_type               = IBV_QPT_RC;
     init_attr.cq_len[UCT_IB_DIR_TX] = config->super.tx_cq_len;
     init_attr.max_rd_atomic         = IBV_DEV_ATTR(&md->super.dev,
                                                    max_qp_rd_atom);
     init_attr.tx_moderation         = config->super.tx_cq_moderation;
+    init_attr.dev_name              = params->mode.device.dev_name;
+    dp_ordering                     = ucs_min(md->dp_ordering_cap_devx.rc,
+                                              UCT_IB_MLX5_DP_ORDERING_OOO_RW);
+
+    status = uct_rc_mlx5_dp_ordering_ooo_init(md, &self->super, dp_ordering, 0,
+                                              &config->rc_mlx5_common, "gga");
+    if (status != UCS_OK) {
+        return status;
+    }
 
     UCS_CLASS_CALL_SUPER_INIT(uct_rc_mlx5_iface_common_t,
                               &uct_gga_mlx5_iface_tl_ops,
@@ -662,14 +813,6 @@ static UCS_CLASS_INIT_FUNC(uct_gga_mlx5_iface_t,
                               &config->super.super, &config->rc_mlx5_common,
                               &init_attr);
 
-    status = uct_rc_mlx5_dp_ordering_ooo_init(
-            &self->super,
-            ucs_min(md->dp_ordering_cap.rc, UCT_IB_MLX5_DP_ORDERING_OOO_RW),
-            &config->rc_mlx5_common, "gga");
-    if (status != UCS_OK) {
-        return status;
-    }
-
     uct_gga_mlx5_iface_disable_rx(&self->super);
 
     config->super.super.fc.enable        = 0; /* FC requires AM capability */
@@ -705,16 +848,23 @@ uct_gga_mlx5_query_tl_devices(uct_md_h md,
 {
     uct_ib_mlx5_md_t *mlx5_md = ucs_derived_of(md, uct_ib_mlx5_md_t);
 
-    if (strcmp(mlx5_md->super.name, UCT_IB_MD_NAME(gga)) ||
-        !ucs_test_all_flags(mlx5_md->flags, UCT_IB_MLX5_MD_FLAG_DEVX |
-                                            UCT_IB_MLX5_MD_FLAG_INDIRECT_XGVMI |
-                                            UCT_IB_MLX5_MD_FLAG_MMO_DMA)) {
+    if (strcmp(mlx5_md->super.name, UCT_IB_MD_NAME(gga))) {
         return UCS_ERR_NO_DEVICE;
     }
 
+    if (!(mlx5_md->super.cap_flags & UCT_MD_FLAG_EXPORTED_MKEY)) {
+        ucs_debug("md %p: cap_flags=0x%" PRIx64 " does not have EXPORTED_MKEY "
+                  "flag", mlx5_md, mlx5_md->super.cap_flags);
+        return UCS_ERR_NO_DEVICE;
+    }
+
+    ucs_assertv(ucs_test_all_flags(mlx5_md->flags, UCT_GGA_MLX5_MD_CAPS),
+                "md %p: flags=0x%x do not have mandatory capabilities 0x%x",
+                mlx5_md, mlx5_md->flags, UCT_GGA_MLX5_MD_CAPS);
+
     return uct_ib_device_query_ports(&mlx5_md->super.dev,
                                      UCT_IB_DEVICE_FLAG_SRQ |
-                                             UCT_IB_DEVICE_FLAG_MLX5_PRM,
+                                     UCT_IB_DEVICE_FLAG_MLX5_PRM,
                                      tl_devices_p, num_tl_devices_p);
 }
 
@@ -725,18 +875,27 @@ UCT_TL_DEFINE_ENTRY(&uct_gga_component, gga_mlx5, uct_gga_mlx5_query_tl_devices,
 
 UCT_SINGLE_TL_INIT(&uct_gga_component, gga_mlx5, ctor,,)
 
+static void uct_ib_mlx5_gga_md_close(uct_md_h md)
+{
+    uct_gga_mlx5_md_t *gga_md = ucs_derived_of(md, uct_gga_mlx5_md_t);
+
+    pthread_mutex_destroy(&gga_md->mem_attach_lock);
+    uct_ib_mlx5_devx_md_close(md);
+}
+
 /* TODO: separate memh since atomic_mr is not relevant for GGA */
 static uct_md_ops_t uct_mlx5_gga_md_ops = {
-    .close              = uct_ib_mlx5_devx_md_close,
-    .query              = uct_ib_mlx5_devx_md_query,
+    .close              = uct_ib_mlx5_gga_md_close,
+    .query              = uct_ib_mlx5_gga_md_query,
     .mem_alloc          = uct_ib_mlx5_devx_device_mem_alloc,
     .mem_free           = uct_ib_mlx5_devx_device_mem_free,
+    .mem_advise         = uct_ib_mem_advise,
     .mem_reg            = uct_ib_mlx5_devx_mem_reg,
     .mem_dereg          = uct_ib_mlx5_devx_mem_dereg,
-    .mem_attach         = uct_ib_mlx5_devx_mem_attach,
-    .mem_advise         = uct_ib_mem_advise,
+    .mem_query          = (uct_md_mem_query_func_t)ucs_empty_function_return_unsupported,
     .mkey_pack          = uct_ib_mlx5_gga_mkey_pack,
-    .detect_memory_type = ucs_empty_function_return_unsupported,
+    .mem_attach         = uct_ib_mlx5_gga_mem_attach,
+    .detect_memory_type = (uct_md_detect_memory_type_func_t)ucs_empty_function_return_unsupported,
 };
 
 static ucs_status_t
@@ -748,7 +907,8 @@ uct_ib_mlx5_gga_md_open(uct_component_t *component, const char *md_name,
     struct ibv_device **ib_device_list, *ib_device;
     int num_devices, fork_init;
     ucs_status_t status;
-    uct_ib_md_t *md;
+    uct_gga_mlx5_md_t *md;
+    int ret;
 
     ucs_trace("opening GGA device %s", md_name);
 
@@ -760,8 +920,7 @@ uct_ib_mlx5_gga_md_open(uct_component_t *component, const char *md_name,
     ib_device_list = ibv_get_device_list(&num_devices);
     if (ib_device_list == NULL) {
         ucs_debug("Failed to get GGA device list, assuming no devices are present");
-        status = UCS_ERR_NO_DEVICE;
-        goto out;
+        return UCS_ERR_NO_DEVICE;
     }
 
     status = uct_ib_get_device_by_name(ib_device_list, num_devices, md_name,
@@ -775,19 +934,30 @@ uct_ib_mlx5_gga_md_open(uct_component_t *component, const char *md_name,
         goto out_free_dev_list;
     }
 
-    status = uct_ib_mlx5_devx_md_open(ib_device, md_config, &md);
+    status = uct_ib_mlx5_devx_md_open_common("gga_mlx5_md_t",
+                                             sizeof(uct_gga_mlx5_md_t),
+                                             ib_device, md_config,
+                                             (uct_ib_md_t**)&md);
     if (status != UCS_OK) {
         goto out_free_dev_list;
     }
 
-    md->super.component = &uct_gga_component;
-    md->super.ops       = &uct_mlx5_gga_md_ops;
-    md->name            = UCT_IB_MD_NAME(gga);
-    md->fork_init       = fork_init;
-    *md_p               = &md->super;
+    md->super.super.super.component = &uct_gga_component;
+    md->super.super.super.ops       = &uct_mlx5_gga_md_ops;
+    md->super.super.name            = UCT_IB_MD_NAME(gga);
+    md->super.super.fork_init       = fork_init;
+
+    ret = pthread_mutex_init(&md->mem_attach_lock, NULL);
+    if (ret != 0) {
+        ucs_error("pthread_mutex_init failed with error: %s", strerror(ret));
+        status = UCS_ERR_IO_ERROR;
+        uct_ib_mlx5_devx_md_close(&md->super.super.super);
+        goto out_free_dev_list;
+    }
+
+    *md_p = &md->super.super.super;
 
 out_free_dev_list:
     ibv_free_device_list(ib_device_list);
-out:
     return status;
 }
