@@ -11,7 +11,9 @@
 #include "tcp.h"
 
 #include <ucs/async/async.h>
+#include <ucs/sys/netlink.h>
 #include <ucs/sys/string.h>
+#include <ucs/sys/sys.h>
 #include <ucs/config/types.h>
 #include <sys/socket.h>
 #include <sys/poll.h>
@@ -137,6 +139,11 @@ static ucs_status_t uct_tcp_iface_get_device_address(uct_iface_h tl_iface,
     dev_addr->flags     = 0;
     dev_addr->sa_family = saddr->sa_family;
 
+    /* Default gateway is not relevant for IPoIB interfaces */
+    if (!ucs_netif_is_ipoib(iface->if_name)) {
+        dev_addr->flags |= UCT_TCP_DEVICE_ADDR_FLAG_ALLOW_DEFAULT_GW;
+    }
+
     if (ucs_sockaddr_is_inaddr_loopback(saddr)) {
         dev_addr->flags |= UCT_TCP_DEVICE_ADDR_FLAG_LOOPBACK;
         memset(pack_ptr, 0, sizeof(uct_iface_local_addr_ns_t));
@@ -201,6 +208,11 @@ uct_tcp_iface_is_reachable_v2(const uct_iface_h tl_iface,
     uct_iface_local_addr_ns_t *local_addr_ns;
     uct_tcp_device_addr_t *tcp_dev_addr;
     int is_local_loopback, is_remote_loopback;
+    struct sockaddr_storage remote_addr;
+    char remote_addr_str[UCS_SOCKADDR_STRING_LEN];
+    unsigned ndev_index;
+    int allow_default_gw;
+    ucs_status_t status;
 
     if (!uct_iface_is_reachable_params_valid(
                 params, UCT_IFACE_IS_REACHABLE_FIELD_DEVICE_ADDR)) {
@@ -236,17 +248,55 @@ uct_tcp_iface_is_reachable_v2(const uct_iface_h tl_iface,
         }
     }
 
-    /* Later connect() call can still fail if the peer is actually unreachable
-     * at UCT/TCP EP creation time */
-    return uct_iface_scope_is_reachable(tl_iface, params);
+    if ((params->field_mask & UCT_IFACE_IS_REACHABLE_FIELD_SCOPE) &&
+        (params->scope == UCT_IFACE_REACHABILITY_SCOPE_DEVICE)) {
+        return uct_iface_scope_is_reachable(tl_iface, params);
+    }
+
+    /* Check if the remote address is routable */
+    status = ucs_ifname_to_index(iface->if_name, &ndev_index);
+    if (status != UCS_OK) {
+        uct_iface_fill_info_str_buf(
+                    params, "failed to get interface index");
+        return 0;
+    }
+
+    remote_addr.ss_family = tcp_dev_addr->sa_family;
+    status = ucs_sockaddr_set_inet_addr((struct sockaddr *)&remote_addr,
+                                        tcp_dev_addr + 1);
+    if (status != UCS_OK) {
+        uct_iface_fill_info_str_buf(
+                    params, "failed to set inet address");
+        return 0;
+    }
+
+    allow_default_gw = !!(tcp_dev_addr->flags &
+                          UCT_TCP_DEVICE_ADDR_FLAG_ALLOW_DEFAULT_GW);
+
+    if (!ucs_netlink_route_exists(ndev_index,
+                                  (const struct sockaddr *)&remote_addr,
+                                  allow_default_gw)) {
+        uct_iface_fill_info_str_buf(
+                    params, "no route to %s",
+                    ucs_sockaddr_str((const struct sockaddr *)&remote_addr,
+                                     remote_addr_str, UCS_SOCKADDR_STRING_LEN));
+        return 0;
+    }
+
+    return 1;
 }
 
 static const char *
 uct_tcp_iface_get_sysfs_path(const char *dev_name, char *path_buffer)
 {
+    const char *sysfs_path = NULL;
     ucs_status_t status;
-    const char *sysfs_path;
-    char lowest_path_buf[PATH_MAX];
+    char *lowest_path_buf;
+
+    status = ucs_string_alloc_path_buffer(&lowest_path_buf, "lowest_path_buf");
+    if (status != UCS_OK) {
+        goto out;
+    }
 
     /* Deep search to find the lowest device sysfs path:
      * 1) For regular device, use regular sysfs form.
@@ -255,11 +305,15 @@ uct_tcp_iface_get_sysfs_path(const char *dev_name, char *path_buffer)
     status = ucs_netif_get_lowest_device_path(dev_name, lowest_path_buf,
                                               PATH_MAX);
     if (status != UCS_OK) {
-        return NULL;
+        goto out_free_lowest_path_buf;
     }
 
     /* 'path_buffer' size is PATH_MAX */
     sysfs_path = ucs_topo_resolve_sysfs_path(lowest_path_buf, path_buffer);
+
+out_free_lowest_path_buf:
+    ucs_free(lowest_path_buf);
+out:
     return sysfs_path;
 }
 
@@ -272,14 +326,19 @@ static ucs_status_t uct_tcp_iface_query(uct_iface_h tl_iface,
     ucs_status_t status;
     int is_default;
     double pci_bw, network_bw, calculated_bw;
-    char path_buffer[PATH_MAX];
+    char *path_buffer;
     const char *sysfs_path;
 
     uct_base_iface_query(&iface->super, attr);
 
     status = uct_tcp_netif_caps(iface->if_name, &attr->latency.c, &network_bw);
     if (status != UCS_OK) {
-        return status;
+        goto out;
+    }
+
+    status = ucs_string_alloc_path_buffer(&path_buffer, "path_buffer");
+    if (status != UCS_OK) {
+        goto out;
     }
 
     sysfs_path             = uct_tcp_iface_get_sysfs_path(iface->if_name, path_buffer);
@@ -340,7 +399,7 @@ static ucs_status_t uct_tcp_iface_query(uct_iface_h tl_iface,
     if (iface->config.prefer_default) {
         status = uct_tcp_netif_is_default(iface->if_name, &is_default);
         if (status != UCS_OK) {
-             return status;
+            goto out_free_path_buffer;
         }
 
         attr->priority    = is_default ? 0 : 1;
@@ -348,7 +407,10 @@ static ucs_status_t uct_tcp_iface_query(uct_iface_h tl_iface,
         attr->priority    = 0;
     }
 
-    return UCS_OK;
+out_free_path_buffer:
+    ucs_free(path_buffer);
+out:
+    return status;
 }
 
 static ucs_status_t uct_tcp_iface_event_fd_get(uct_iface_h tl_iface, int *fd_p)
@@ -500,7 +562,7 @@ static uct_iface_ops_t uct_tcp_iface_ops = {
     .iface_progress_disable   = uct_base_iface_progress_disable,
     .iface_progress           = uct_tcp_iface_progress,
     .iface_event_fd_get       = uct_tcp_iface_event_fd_get,
-    .iface_event_arm          = ucs_empty_function_return_success,
+    .iface_event_arm          = (uct_iface_event_arm_func_t)ucs_empty_function_return_success,
     .iface_close              = UCS_CLASS_DELETE_FUNC_NAME(uct_tcp_iface_t),
     .iface_query              = uct_tcp_iface_query,
     .iface_get_address        = uct_tcp_iface_get_address,
@@ -615,13 +677,16 @@ static ucs_mpool_ops_t uct_tcp_mpool_ops = {
 };
 
 static uct_iface_internal_ops_t uct_tcp_iface_internal_ops = {
-    .iface_estimate_perf   = uct_base_iface_estimate_perf,
-    .iface_vfs_refresh     = (uct_iface_vfs_refresh_func_t)ucs_empty_function,
-    .ep_query              = (uct_ep_query_func_t)ucs_empty_function_return_unsupported,
-    .ep_invalidate         = (uct_ep_invalidate_func_t)ucs_empty_function_return_unsupported,
-    .ep_connect_to_ep_v2   = uct_tcp_ep_connect_to_ep_v2,
-    .iface_is_reachable_v2 = uct_tcp_iface_is_reachable_v2,
-    .ep_is_connected       = uct_tcp_ep_is_connected
+    .iface_query_v2         = uct_iface_base_query_v2,
+    .iface_estimate_perf    = uct_base_iface_estimate_perf,
+    .iface_vfs_refresh      = (uct_iface_vfs_refresh_func_t)ucs_empty_function,
+    .iface_mem_element_pack = (uct_iface_mem_element_pack_func_t)ucs_empty_function_return_unsupported,
+    .ep_query               = (uct_ep_query_func_t)ucs_empty_function_return_unsupported,
+    .ep_invalidate          = (uct_ep_invalidate_func_t)ucs_empty_function_return_unsupported,
+    .ep_connect_to_ep_v2    = uct_tcp_ep_connect_to_ep_v2,
+    .iface_is_reachable_v2  = uct_tcp_iface_is_reachable_v2,
+    .ep_is_connected        = uct_tcp_ep_is_connected,
+    .ep_get_device_ep       = (uct_ep_get_device_ep_func_t)ucs_empty_function_return_unsupported
 };
 
 static UCS_CLASS_INIT_FUNC(uct_tcp_iface_t, uct_md_h md, uct_worker_h worker,
@@ -883,13 +948,25 @@ static UCS_CLASS_DEFINE_NEW_FUNC(uct_tcp_iface_t, uct_iface_t, uct_md_h,
 
 static int uct_tcp_is_bridge(const char *if_name)
 {
-    char path[PATH_MAX];
+    char *path;
+    int ret;
     struct stat st;
+    ucs_status_t status;
 
-    ucs_snprintf_safe(path, PATH_MAX, UCT_TCP_IFACE_NETDEV_DIR "/%s/bridge",
-                      if_name);
+    status = ucs_string_alloc_formatted_path(&path, "path",
+                                             UCT_TCP_IFACE_NETDEV_DIR
+                                             "/%s/bridge",
+                                             if_name);
+    if (status != UCS_OK) {
+        ret = 0;
+        goto out;
+    }
 
-    return (stat(path, &st) == 0) && S_ISDIR(st.st_mode);
+    ret = (stat(path, &st) == 0) && S_ISDIR(st.st_mode);
+
+    ucs_free(path);
+out:
+    return ret;
 }
 
 ucs_status_t uct_tcp_query_devices(uct_md_h md,
@@ -904,7 +981,7 @@ ucs_status_t uct_tcp_query_devices(uct_md_h md,
     int is_active, i, n;
     ucs_status_t status;
     const char *sysfs_path;
-    char path_buffer[PATH_MAX];
+    char *path_buffer;
     ucs_sys_device_t sys_dev;
 
     n = scandir(UCT_TCP_IFACE_NETDEV_DIR, &entries, NULL, alphasort);
@@ -916,6 +993,12 @@ ucs_status_t uct_tcp_query_devices(uct_md_h md,
 
     devices     = NULL;
     num_devices = 0;
+
+    status = ucs_string_alloc_path_buffer(&path_buffer, "path_buffer");
+    if (status != UCS_OK) {
+        goto out;
+    }
+
     ucs_carray_for_each(entry, entries, n) {
         /* According to the sysfs(5) manual page, all of entries
          * has to be a symbolic link representing one of the real
@@ -978,6 +1061,7 @@ out_release:
     }
 
     free(entries);
+    ucs_free(path_buffer);
 out:
     return status;
 }
