@@ -33,6 +33,7 @@
 #include <dirent.h>
 #include <stdint.h>
 #include <ifaddrs.h>
+#include <sys/resource.h>
 
 
 #ifndef UINT16_MAX
@@ -338,6 +339,7 @@ std::string get_rdmacm_netdev(const char *ifa_name);
  */
 bool is_rdmacm_netdev(const char *ifa_name);
 
+bool is_aws();
 
 /**
  * Get an available port on the host.
@@ -1054,6 +1056,13 @@ const std::vector<std::vector<ucs_memory_type_t> >& supported_mem_type_pairs();
  */
 void skip_on_address_sanitizer();
 
+
+/**
+ * Returns whether to skip HW TM offload tests
+ */
+bool skip_hw_tm_offload();
+
+
 /**
  * Class for mocking C function pointers
  * The class is used to mock C function pointers, by replacing the original
@@ -1124,6 +1133,57 @@ private:
     std::unordered_map<func*, std::unique_ptr<mock_func>> m_mock_map;
 };
 
+#define UCS_MOCK_ORIG_FUNC(_mock, _orig_ptr, ...) \
+{ \
+    ucs_status_t _status = _mock.orig_func(_orig_ptr, ##__VA_ARGS__); \
+    if (_status != UCS_OK) { \
+        return _status; \
+    } \
+}
+
+// Used to manipulate and restore rlimits
+class rlimit_setter {
+public:
+    rlimit_setter(int resource, rlim_t new_limit) :
+        m_resource(resource), m_limit(0), m_orig_limit({})
+    {
+        if (getrlimit(m_resource, &m_orig_limit)) {
+            UCS_TEST_MESSAGE << "could not get rlimit " << resource
+                             << ", err: " << strerror(errno);
+            return;
+        }
+
+        struct rlimit limit = {
+            .rlim_cur = std::min(new_limit, m_orig_limit.rlim_max),
+            .rlim_max = m_orig_limit.rlim_max
+        };
+        if (setrlimit(m_resource, &limit) == 0) {
+            m_limit = limit.rlim_cur;
+        } else {
+            m_limit = m_orig_limit.rlim_cur;
+            UCS_TEST_MESSAGE << "could not set rlimit " << resource << " to "
+                             << limit.rlim_cur << ", err: " << strerror(errno);
+        }
+    }
+
+    ~rlimit_setter()
+    {
+        if (m_orig_limit.rlim_cur != m_limit) {
+            (void)setrlimit(m_resource, &m_orig_limit);
+        }
+    }
+
+    rlim_t limit() const
+    {
+        return m_limit;
+    }
+
+private:
+    const int     m_resource;
+    rlim_t        m_limit;
+    struct rlimit m_orig_limit;
+};
+
 } // ucs
 
 #endif /* UCS_TEST_HELPERS_H */
