@@ -246,7 +246,8 @@ public:
 
         ucp_ep_purge_lanes(ep, purge_enqueue_cb, this);
         void *request = ucp_ep_flush_internal(ep, 0, &param, NULL, flushed_cb,
-                                              "ep_restart");
+                                              "ep_restart",
+                                              UCT_FLUSH_FLAG_LOCAL);
 
         ASSERT_FALSE(UCS_PTR_IS_ERR(request));
         if (request != NULL) {
@@ -374,10 +375,9 @@ public:
     }
 
     void reset_protocol(operation_t op, bool sync = false,
-                        unsigned reqs_count = 1000)
+                        unsigned reqs_count = 1000,
+                        size_t msg_size = UCS_KBYTE * 70)
     {
-        static const size_t msg_size = UCS_KBYTE * 70;
-
         for (int i = 0; i < reqs_count; ++i) {
             mapped_buffer *rbuf = new mapped_buffer(msg_size, receiver());
             rbuf->memset(0);
@@ -390,6 +390,7 @@ public:
         std::vector<void*> reqs;
         send_requests(1, reqs, op, sync);
         wait_reqs(op, reqs);
+        flush_workers();
 
         /* Send all messages */
         send_requests(reqs_count, reqs, op, sync);
@@ -468,6 +469,8 @@ protected:
                                                                            NULL;
     }
 
+    const ucp_request_t *get_rndv_request(const void *ureq);
+
     std::vector<std::vector<uint8_t>>           m_sbufs;
     std::vector<std::unique_ptr<mapped_buffer>> m_rbufs;
     std::vector<ucs::handle<ucp_rkey_h>>        m_rkeys;
@@ -578,22 +581,23 @@ UCP_INSTANTIATE_TEST_CASE(test_proto_reset)
 /* The following tests require ENABLE_DEBUG_DATA flag in order to access
  * req->recv.proto_rndv_request, which is only present with this flag. */
 #if ENABLE_DEBUG_DATA
+const ucp_request_t *test_proto_reset::get_rndv_request(const void *ureq)
+{
+    auto req = get_request(ureq, false);
+    if ((req == NULL) || (req->recv.proto_rndv_request == NULL)) {
+        return NULL;
+    }
+
+    return req->recv.proto_rndv_request;
+}
+
 class test_proto_reset_rndv_get : public test_proto_reset {
 protected:
     void wait_and_restart(const std::vector<void*> &reqs) override
     {
         wait_any(reqs, [this](const void *ureq) {
-            const ucp_request_t *req = get_request(ureq, false);
-            if (req == NULL) {
-                return false;
-            }
-
-            const ucp_request_t *rndv_req = req->recv.proto_rndv_request;
-            if (rndv_req == NULL) {
-                return false;
-            }
-
-            return is_request_in_the_middle(rndv_req);
+            auto rndv_req = get_rndv_request(ureq);
+            return (rndv_req != NULL) && is_request_in_the_middle(rndv_req);
         });
 
         restart(receiver().ep());
@@ -613,6 +617,40 @@ UCS_TEST_P(test_proto_reset_rndv_get, rndv_get, "RNDV_THRESH=0",
 
 UCP_INSTANTIATE_TEST_CASE(test_proto_reset_rndv_get)
 
+class test_proto_reset_rkey_ptr : public test_proto_reset {
+protected:
+    void wait_and_restart(const std::vector<void*> &reqs) override
+    {
+        wait_any(reqs, [this](const void *ureq) {
+            auto rndv_req = get_rndv_request(ureq);
+            return (rndv_req != NULL) &&
+                   (rndv_req->send.state.completed_size > 0);
+        });
+
+        ucs_queue_iter_t iter;
+        ucp_request_t *req;
+
+        ucs_queue_for_each_safe(req, iter, &receiver().worker()->rkey_ptr_reqs,
+                                send.rndv.rkey_ptr.queue_elem) {
+            m_pending.push_back(req);
+        }
+
+        restart(receiver().ep());
+    }
+};
+
+UCS_TEST_P(test_proto_reset_rkey_ptr, rkey_ptr, "RNDV_THRESH=0",
+           "RKEY_PTR_SEG_SIZE=1024")
+{
+    if (!has_resource(sender(), "xpmem")) {
+        UCS_TEST_SKIP_R("xpmem must be present for rkey_ptr protocol");
+    }
+
+    reset_protocol(TAG, false, 1000, UCS_KBYTE * 20);
+}
+
+UCP_INSTANTIATE_TEST_CASE_TLS(test_proto_reset_rkey_ptr, shm_ib, "shm,ib")
+
 /* This test is intended to check resetting request during ATP phase for
  * RNDV_PUT protocol.
  * We use uct hooks on some UCT level functions in order to simulate the
@@ -703,8 +741,18 @@ protected:
                     UCP_PROTO_RNDV_PUT_STAGE_FENCED_ATP);
         });
 
-        restart(sender().ep());
+
+       /*
+        * Restore the pending_add callbacks, as some transport need to be able
+        * to add ep flush request to the pending queue. For instance RDNV put
+        * and fence scheme will block ep_flush on SRD transport.
+        *
+        * Purge must be done with test callbacks first.
+        */
+        ucp_ep_purge_lanes(sender().ep(), purge_enqueue_cb, this);
         restore_uct_cbs();
+
+        restart(sender().ep());
     }
 
     static unsigned m_atp_count;
