@@ -31,7 +31,8 @@ public:
         VARIANT_RNDV_AM_BCOPY,
         VARIANT_RNDV_AM_ZCOPY,
         VARIANT_SEND_NBR,
-        VARIANT_PROTO_V1
+        VARIANT_PROTO_V1,
+        VARIANT_CONNECT_ALL_TO_ALL
     };
 
     test_ucp_tag_xfer() {
@@ -61,6 +62,8 @@ public:
             modify_config("ZCOPY_THRESH", "0");
         } else if (get_variant_value() == VARIANT_PROTO_V1) {
             modify_config("PROTO_ENABLE", "n");
+        } else if (get_variant_value() == VARIANT_CONNECT_ALL_TO_ALL) {
+            modify_config("CONNECT_ALL_TO_ALL", "y");
         }
 
         /* Init number of lanes according to test requirement
@@ -98,6 +101,8 @@ public:
             add_variant_with_value(variants, get_ctx_params(), VARIANT_PROTO_V1,
                                    "proto_v1");
         }
+        add_variant_with_value(variants, get_ctx_params(),
+                               VARIANT_CONNECT_ALL_TO_ALL, "connect_all_to_all");
     }
 
     virtual ucp_ep_params_t get_ep_params() {
@@ -1094,7 +1099,7 @@ public:
 
     bool has_rma_zcopy() {
         return has_transport("rc_v") || has_transport("rc_x") ||
-               has_transport("dc_x") ||
+               has_transport("dc_x") || has_transport("srd") ||
                (ucp_context_find_tl_md(receiver().ucph(), "cma")  != NULL) ||
                (ucp_context_find_tl_md(receiver().ucph(), "knem") != NULL);
     }
@@ -1245,20 +1250,30 @@ UCS_TEST_P(multi_rail_max, max_lanes, "IB_NUM_PATHS?=64", "TM_SW_RNDV=y",
     receiver().connect(&sender(), get_ep_params());
     test_run_xfer(true, true, true, true, false);
 
-    EXPECT_EQ(num_lanes(), ucp_ep_num_lanes(sender().ep()));
-    EXPECT_EQ(num_lanes(), ucp_ep_num_lanes(receiver().ep()));
+    auto ep_num_lanes = ucp_ep_num_lanes(sender().ep());
+    ASSERT_EQ(ep_num_lanes, ucp_ep_num_lanes(receiver().ep()));
+
+    if (is_proto_enabled()) {
+        EXPECT_EQ(num_lanes(), ep_num_lanes);
+    }
 
     size_t bytes_sent = 0;
-    for (ucp_lane_index_t lane = 0; lane < num_lanes(); ++lane) {
+    unsigned num_used_lanes = 0;
+    for (ucp_lane_index_t lane = 0; lane < ep_num_lanes; ++lane) {
         size_t sender_tx   = get_bytes_sent(sender().ep(), lane);
         size_t receiver_tx = get_bytes_sent(receiver().ep(), lane);
         UCS_TEST_MESSAGE << "lane[" << static_cast<int>(lane) << "] : "
                          << "sender " << sender_tx << " receiver " << receiver_tx;
-
-        EXPECT_GT(sender_tx + receiver_tx, 0);
+        if ((sender_tx > 0) || (receiver_tx > 0)) {
+            ++num_used_lanes;
+        }
         bytes_sent += sender_tx + receiver_tx;
     }
 
+    /* One lane could be reserved for tag offload and not selected for AM/RMA
+       bandwidth lane */
+    EXPECT_GE(num_used_lanes, ep_num_lanes - 1);
+
     EXPECT_GE(bytes_sent, get_msg_size());
 }
 
