@@ -619,7 +619,13 @@ unsigned ucp_test::mt_num_threads()
 #if _OPENMP && ENABLE_MT
     /* Assume each thread can create two workers (sender and receiver entity),
        and each worker can open up to 64 files */
-    return std::min(omp_get_max_threads(), ucs_sys_max_open_files() / (64 * 2));
+    unsigned num = std::min(omp_get_max_threads(),
+                            ucs_sys_max_open_files() / (64 * 2));
+    if (ucs::is_aws()) {
+        num = std::min(num, 64u);
+    }
+
+    return num;
 #else
     return 1;
 #endif
@@ -1208,6 +1214,12 @@ bool ucp_test_base::entity::is_rndv_put_ppln_supported() const
     return false;
 }
 
+bool ucp_test_base::entity::is_rndv_supported() const
+{
+    const auto config = ucp_ep_config(ep());
+    return config->key.rma_bw_lanes[0] != UCP_NULL_LANE;
+}
+
 bool ucp_test_base::entity::is_conn_reqs_queue_empty() const
 {
     return m_conn_reqs.empty();
@@ -1286,6 +1298,11 @@ ucp_mem_h ucp_test::mapped_buffer::memh() const
     return m_memh;
 }
 
+ucp_worker_h ucp_test::mapped_buffer::worker() const
+{
+    return m_entity.worker();
+}
+
 void test_ucp_context::get_test_variants(std::vector<ucp_test_variant> &variants)
 {
     add_variant(variants, UCP_FEATURE_TAG | UCP_FEATURE_WAKEUP);
