@@ -28,6 +28,7 @@ std::string resource::name() const {
 }
 
 resource::resource() : component(NULL), dev_type(UCT_DEVICE_TYPE_LAST),
+                       sys_device(UCS_SYS_DEVICE_ID_UNKNOWN),
                        variant(DEFAULT_VARIANT)
 {
     CPU_ZERO(&local_cpus);
@@ -36,10 +37,11 @@ resource::resource() : component(NULL), dev_type(UCT_DEVICE_TYPE_LAST),
 resource::resource(uct_component_h component, const std::string& component_name,
                    const std::string& md_name, const ucs_cpu_set_t& local_cpus,
                    const std::string& tl_name, const std::string& dev_name,
-                   uct_device_type_t dev_type) :
+                   uct_device_type_t dev_type, ucs_sys_device_t sys_dev) :
                    component(component), component_name(component_name),
                    md_name(md_name), local_cpus(local_cpus), tl_name(tl_name),
                    dev_name(dev_name), dev_type(dev_type),
+                   sys_device(sys_dev),
                    variant(DEFAULT_VARIANT)
 {
 }
@@ -56,6 +58,7 @@ resource::resource(uct_component_h component,
                    tl_name(tl_resource.tl_name),
                    dev_name(tl_resource.dev_name),
                    dev_type(tl_resource.dev_type),
+                   sys_device(tl_resource.sys_device),
                    variant(DEFAULT_VARIANT)
 {
 }
@@ -110,12 +113,15 @@ resource_speed::resource_speed(uct_component_h component,
 std::vector<uct_test_base::md_resource> uct_test_base::enum_md_resources() {
 
     static std::vector<uct_test::md_resource> all_md_resources;
+    static bool populated = false;
 
-    if (all_md_resources.empty()) {
+    if (!populated) {
         uct_component_h *uct_components;
         unsigned num_components;
         ucs_status_t status;
 
+        const char *str = getenv("GTEST_MAX_COMP_RESOURCES");
+
         status = uct_query_components(&uct_components, &num_components);
         ASSERT_UCS_OK(status);
 
@@ -147,12 +153,19 @@ std::vector<uct_test_base::md_resource> uct_test_base::enum_md_resources() {
                                          &component_attr_resouces);
             ASSERT_UCS_OK(status);
 
+            int md_resource_count = md_rsc.cmpt_attr.md_resource_count;
+            if (str != NULL) {
+                md_resource_count = ucs_min(md_resource_count, atoi(str));
+            }
+
             for (unsigned md_index = 0;
-                 md_index < md_rsc.cmpt_attr.md_resource_count; ++md_index) {
+                 md_index < md_resource_count; ++md_index) {
                 md_rsc.rsc_desc = md_resources[md_index];
                 all_md_resources.push_back(md_rsc);
             }
         }
+
+        populated = true;
     }
 
     return all_md_resources;
@@ -248,7 +261,7 @@ void uct_test::set_interface_rscs(uct_component_h cmpt, const char *cmpt_name,
     for (i = 0; i < 3; i++) {
         resource rsc(cmpt, std::string(cmpt_name), std::string(md_name),
                      local_cpus, "", std::string(ifa->ifa_name),
-                     UCT_DEVICE_TYPE_NET);
+                     UCT_DEVICE_TYPE_NET, UCS_SYS_DEVICE_ID_UNKNOWN);
         bool init_src_addr = (i == 1);
 
         if (i < 2) {
@@ -596,6 +609,10 @@ bool uct_test::has_ud() const {
     return (has_transport("ud_verbs") || has_transport("ud_mlx5"));
 }
 
+bool uct_test::has_srd() const {
+    return has_transport("srd");
+}
+
 bool uct_test::has_rc() const {
     return (has_transport("rc_verbs") || has_transport("rc_mlx5"));
 }
@@ -1030,7 +1047,14 @@ void uct_test::entity::rkey_unpack(const uct_allocated_memory_t *mem,
             UCS_TEST_ABORT("Failed to allocate rkey buffer");
         }
 
-        ucs_status_t status = uct_md_mkey_pack(m_md, mem->memh, rkey_buffer);
+        uct_md_mkey_pack_params_t params = {
+            .field_mask = 0
+        };
+
+        ucs_status_t status = uct_md_mkey_pack_v2(m_md, mem->memh,
+                                                  mem->address,
+                                                  mem->length,
+                                                  &params, rkey_buffer);
         ASSERT_UCS_OK(status);
 
         status = uct_rkey_unpack(m_resource.component, rkey_buffer,
@@ -1452,12 +1476,13 @@ uct_test::mapped_buffer::mapped_buffer(size_t size,
     if ((mem_type == UCS_MEMORY_TYPE_HOST) || (mem_type == UCS_MEMORY_TYPE_RDMA)) {
         m_entity.mem_alloc(alloc_size, mem_flags, &m_mem, mem_type, num_retries);
     } else {
-        m_mem.method   = UCT_ALLOC_METHOD_LAST;
-        m_mem.address  = mem_buffer::allocate(alloc_size, mem_type);
-        m_mem.length   = alloc_size;
-        m_mem.mem_type = mem_type;
-        m_mem.memh     = UCT_MEM_HANDLE_NULL;
-        m_mem.md       = NULL;
+        m_mem.method     = UCT_ALLOC_METHOD_LAST;
+        m_mem.address    = mem_buffer::allocate(alloc_size, mem_type);
+        m_mem.length     = alloc_size;
+        m_mem.mem_type   = mem_type;
+        m_mem.memh       = UCT_MEM_HANDLE_NULL;
+        m_mem.md         = NULL;
+        m_mem.sys_device = UCS_SYS_DEVICE_ID_UNKNOWN;
         m_entity.mem_type_reg(&m_mem, mem_flags);
     }
 
